/*******************************************************************
 *******************************************************************

 Class: FTT-Ethernet                                  
 Type:  FTT Ethernet application example
 Platform: Shark
 Name: eg_em_m.c
 
 By: Paulo Pedreiras, Ricardo Leite
 Date: 2003-Jun
 Description: Application using the Elastic Message Model + EDF
              Implements a basic operator console allowing
                to dynamically changing the communication requirements
              Master component

 Log: (Date,Author,description)Comments:

 ********************************************************************
 ********************************************************************/

#include "fttm1_0.h"

/********************************/
/* main()                       */
/********************************/

int main(int argc, char **argv)
{
  /* Auxiliary variable used to append messages to the SRT */
  SRDB_SRT_mesgtype SRT_aux_var; 

 /* Variables required by the user interface (text mode) */
  int CR = 13;        /* ASCII code for RETURN */
  int CMD_MAXLEN=80; /* Maximum chars in a commad line */
  int car,nc=0;
  char cmd[CMD_MAXLEN];
  
  int m_id,m_size,m_nomp,m_minp, m_maxp,m_ecoef,m_ddl,m_initph; /* Message attributes */
  
  int stat; /* Status reported by FTT function invocations */

  
 /* FTT global vars initialization */
  EC_LEN=10000; /* EC length (us) */
  tm_seqn=0; /* Initial value for Trigger Message sequnce number */   
 

    /***********************/
   /* Init the ftt system */
  /***********************/
  ftt_minit();


  /**************************/
  /* Set-up the message set */
  /**************************/
  /* Synchronous message set*/
  
  cprintf("\n\n Elastic message model demo !\n\n");
  cprintf("\n Building synchronous message set ...");

#ifdef ELASTIC_TM
  SET_SMESG_PROP(&SRT_aux_var,1,1024,1,1, 5, 2, 5,0); /* id, size, period, period_min, period_max,elast. coef,deadline, init */ 
  SRDB_SRT_addmesg(&SRT_aux_var);

  SET_SMESG_PROP(&SRT_aux_var,2,1024,1,1,10, 5,10,0); 
  SRDB_SRT_addmesg(&SRT_aux_var);

  SET_SMESG_PROP(&SRT_aux_var,3,1024,1,1,15,10,15,0); 
  SRDB_SRT_addmesg(&SRT_aux_var);

  SET_SMESG_PROP(&SRT_aux_var,4,1024,1,1,30,50,30,0); 
  SRDB_SRT_addmesg(&SRT_aux_var);
#else

  SET_SMESG_PROP(&SRT_aux_var,1,1024,1,1,0); /* id, size, period, deadline, init */ 
  SRDB_SRT_addmesg(&SRT_aux_var);
  
  SET_SMESG_PROP(&SRT_aux_var,2,1024,1,1,0); 
  SRDB_SRT_addmesg(&SRT_aux_var);

  SET_SMESG_PROP(&SRT_aux_var,3,1024,1,1,0); 
  SRDB_SRT_addmesg(&SRT_aux_var);
  
#endif

  cprintf(" Finished building message set!");

  /* Print the initial message set */
  /* Synchronous messages */
  cprintf("\n Message set:");
  SRDB_SRT_printmesg();
  cprintf("\n Any key to continue");
  keyb_getchar();


    /*****************************************/
   /* Messages configured. Start the system */
  /*****************************************/
  ftt_mstart();

  /* User interface for message set management */
  cprintf("\n\n Management console activated. H <ENTER> for Help!\n\n");

  while(1)
    {
      car = keyb_getchar();
      if(car >= 'a' && car <='z') car=car-('a'-'A');
      if(car != CR && car != ' ' && (car < 'A' || car > 'Z') && (car < '0' || car > '9')) continue;
      if(car == CR)
	{
	  int ind=0;
	  /* Command entered. Process it */
	  /* printf("\n |%s|\n ",cmd);*/
	  while(cmd[ind]==' ')
	    ind++;
	  switch(cmd[ind])
	    {
	    case 'H':
	      cprintf("\n Management console: available commands:");
#ifdef ELASTIC_TM
	      cprintf("\n  A ID SZ NomP MinP MaxP ElCoef Ddln IPh  : Add a new message ");
#else
	      cprintf("\n  A ID SZ P D : Add a new message ");
#endif
	      cprintf("\n  D ID      : Delete message identified by ID");
	      cprintf("\n  P : Print the message table");
	      cprintf("\n  S : Statistics" );
	      cprintf("\n  X : eXit\n");
	      break;
	      
	    case 'A':
#ifdef ELASTIC_TM
	      sscanf(&cmd[ind+1],"%d %d %d %d %d %d %d %d",&m_id,&m_size,&m_nomp,&m_minp,&m_maxp,&m_ecoef,&m_ddl,&m_initph);
	      SET_SMESG_PROP(&SRT_aux_var,m_id,m_size,m_nomp,m_minp, m_maxp,m_ecoef,m_ddl,m_initph);  
#else
	      sscanf(&cmd[ind+1],"%d %d %d %d",&m_id,&m_size,&m_nomp,&m_ddl);
	      SET_SMESG_PROP(&SRT_aux_var,m_id,m_size,m_nomp,m_ddl,0);  

#endif
	      stat=SRDB_SRT_addmesg(&SRT_aux_var);
	      /*      - error code: 0 : OK                                 */
	      /*                   -1 : Table full (table->nvars = MAX_VAR)*/
	      /*                   -2 : Set unschedulable                  */
	      /*                   -3 : Invalid message properties         */
	      /*                   -4 : Invalid message type               */
	      /*                   -5 : Invalid ID (duplicated)            */
	      switch(stat)
		{
		case 0:
		  cprintf("\n Message %d added sucessfully!\n\n",m_id);
		  break;
		case -1:
		  cprintf("\n Error! Message table full.\n\n");
		  break;
		case -2:
		  cprintf("\n Error! Message set unschedulable.\n\n");
		  break;
		case -3:
		  cprintf("\n Error! Invalid message attributes.\n\n");
		  break;
		case -4:
		  cprintf("\n Error! Invalid message type.\n\n");
		  break;
		case -5:
		  cprintf("\n Error! Duplicated message ID.\n\n");
		  break;
		}
	      break;
	      
	    case 'D':
	      sscanf(&cmd[ind+1],"%d",&m_id);
	      stat=SRDB_SRT_delmesg(m_id);
	      /* error code: 0 : OK        */
	      /*            -1 : Table empty*/
	      /*            -2 : Invalid ID*/
	      switch(stat)
		{
		case 0:
		  cprintf("\n Message %d removed sucessfully!\n\n",m_id);
		  break;
		case -1:
		  cprintf("\n Error! Message table empty.\n\n");
		  break;
		case -2:
		  cprintf("\n Error! Message not found.\n\n");
		  break;
		}
	      	      
	      break;
	      
	    case 'P':
	      cprintf("\n Message set:");
	      SRDB_SRT_printmesg();
	      cprintf("\n\n");
	      
	      break;
	      
	    case 'S':
	      {
		double Ci,U_nom=0,U_min=0,U_act=0,U_lub=0;
		int i,nvars = SRDB_SRT_mesgtable.sm_nvars;


		SRDB_SRT_GetUf(&SRDB_SRT_mesgtable,&U_act,&U_lub);
#ifdef ELASTIC_TM
		/* Get nominal and minimum utilization factors for the message set */
		for(i=0;i<nvars;i++)
		  {
		    Ci=(double)SRDB_SRT_mesgtable.sm_message[i].sm_ssize*SLOTV_US;
		    U_nom+=Ci/(SRDB_SRT_mesgtable.sm_message[i].sm_period_nom*EC_LEN);
		    U_min+=Ci/(SRDB_SRT_mesgtable.sm_message[i].sm_period_max*EC_LEN);
		  }
#endif

		cprintf("\n Synchronous Message Set:");
		cprintf("\n     Number of messages ............ = %d", nvars);
		cprintf("\n     EC length ..................... = %lu us",EC_LEN);
		cprintf("\n     Synchronous window share ...... = %4.2f (%%)",SWIN_SHARE*100);
		cprintf("\n     Guaranteed schedulability limit = %4.2f (%%)",U_lub*100);
		cprintf("\n     Current utilization ........... = %4.2f (%%)",U_act*100);
#ifdef ELASTIC_TM
		cprintf("\n     Nominal utilization ........... = %4.2f (%%)",U_nom*100);
		cprintf("\n     Minimum utilization ........... = %4.2f (%%)",U_min*100);
#endif
		cprintf("\n");
	      }
	      break;
	      
	    case 'X':
	      sys_end();
	      
	    default:
	      if(nc != 0)
		cprintf("\n Unknown command! Type H <ENTER> for help\n\n");
	      
	    }
	  
	  
	  /* Reset cmd */
	  nc=0;
	  cmd[0]=0;
	}
      else
	{
	  cprintf("%c",car);
	  if(nc<CMD_MAXLEN-1)
	    {
	      cmd[nc++]=(char)car;
	      cmd[nc]=0;
	    }
	}
    }

  
  /* Main task ends but system does not shutdown since there are active tasks */
  return 0;
  
}
