/*******************************************************************
 *******************************************************************

 Class: FTT-Ethernet                                  
 Type:  FTT Ethernet example application
 Platform: Shark
 Name: eg_em_s.c
 
 By: Paulo Pedreiras
 Date: 2003-Ago-4
 Description: Two slave nodes, selected by ID {1,2} in the command line
              Configures a set of synchronous messages
              Uses Earliest Deadline First + Elastic Task Model (at the master node)
              Slave component

 Log: (Date,Author,description)Comments:

 ********************************************************************
 ********************************************************************/


#include "ftts1_0.h"

int stationid; /* Selects between station 1 or 2 */

    /*****************************************************/
   /*****************************************************/
  /*           Main                                    */
 /*****************************************************/
/*****************************************************/

int main(int argc, char **argv)
{
  
  /* General variables */
  NRDB_mesgtype aux_var;
  signed char vstatus;

  //set_exchandler_grx(); /* After execption prints PID ... */


  /*   FTT global vars init      */
  EC_LEN=10000; /* EC length (us) */
  
  /* Check input arguments */
  if(argc != 2)
    {
      cprintf("\n\n Invalid number of arguments");
      cprintf("\n Usage: eg_em_s ID   (ID={1,2})");
      sys_end();
    }

  stationid=atoi(argv[1]);
  if(stationid != 1 && stationid != 2)
    {
      cprintf("\n\n Invalid station ID!");
      cprintf("\n Usage: eg_em_s ID   (ID={1,2})");
      sys_end();
    }


  /* Init the ftt system */
  ftt_ninit();


    /**************************************/
   /* Set-up the message set             */
  /**************************************/

  if(stationid == 1)
    {
      
      /* STA1 : Message set */
      
      SET_MESG_PROP(&aux_var,1,1024,5,SDATA_MESG_ID,PRODUCER,0); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
      SET_MESG_PROP(&aux_var,2,1024,10,SDATA_MESG_ID,PRODUCER,0); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
    }

  if(stationid == 2)
    {     
      /* STA2 : Message set */
      
      SET_MESG_PROP(&aux_var,3,1024,15,SDATA_MESG_ID,PRODUCER,0); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
      SET_MESG_PROP(&aux_var,4,1024,30,SDATA_MESG_ID,PRODUCER,0); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
    }


  /* Print station and message set information */
  cprintf("\n-----------------------------");
  if(stationid == 1)
    cprintf("\n    Station 1 ");

  if(stationid == 2)
    cprintf("\n    Station 2 ");
  cprintf("\n-----------------------------");
  
  /* Print additional information (message tables) */
  cprintf("\n Any key to see the message tables");
  keyb_getchar();
  
  /* Print message tables */
  NRDB_printmesg();
  
  cprintf("\n Any key to continue");
  keyb_getchar();


  /* Start the ftt system */
  ftt_nstart();
  
  cprintf("\n System running...");


  return 0;
  
}

