/*******************************************************************
 ******************************************************************/
/*!	\file	fttm1_0.c
	\brief	Functions code  related with the FTT-Ethernet\n
		    Master implementation

 Functions code  related with the FTT-Ethernet\n
		    Master implementation

 Class: FTT-Ethernet\n
 Type: FTT Ethernet Master master code\n
 Platform: Shark\n
 Name: fttm1_0.c\n\n
 
 By: Paulo Pedreiras\n
 Date: 2003-Jul-31\n
 */
 
 /*******************************************************************
 ********************************************************************/


  /********************************/
 /* Includes 			         */
/********************************/

#include "fttm1_0.h"


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Function Implementation >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/


    /*****************************************************/
   /*****************************************************/
  /*     CBS handling functions                        */
 /*****************************************************/
/*****************************************************/

#ifdef CBS_SERVER

void CBS_init(CBS_list_type *CBS_list)
/********************************************************/
/* This function initializes the CBS data structures    */
/********************************************************/
{
  CBS_list->ns=0;           /* No CBS servers defined */
}


int CBS_getUcbs(CBS_list_type *CBS_list,double * Ucbs)	///< tetse telkhlkj
/**********************************************************/
/* This function returns the utilization of the           */
/*   CBS servers currently defined                        */
/* Inputs:                                                */
/*     *CBS_list: list of servers                         */
/*                                                        */
/* Outputs:                                               */
/*     Ucbs: utilization of the servers                   */      
/* Returns:                                               */
/*     0: Sucess                                          */
/*    -1: Empty CBS list                                  */
/**********************************************************/
{

  WORD ns;
  int i;
  
  /* Var init */
  *Ucbs=0;
  ns=CBS_list->ns;

  /* Check if any servers defined */
  if(ns == 0)
    return -1;

  /* Compute utilizations of CBSs ... */
  for(i=0;i<ns;i++)
    (*Ucbs) += (double)(CBS_list->cbs[i].Q*8*BIT_TIME_US / (CBS_list->cbs[i].T*EC_LEN) );

  return 0;

}


int CBS_addserver(CBS_list_type *CBS_list, WORD id, WORD Q, WORD T )
/**********************************************************/
/* This function add a CBS server to the CBS servers list */
/* Inputs:                                                */
/*      id: message served by the CBS                     */
/*       Q: CBS maximum budget                            */
/*       T: CBS period                                    */
/* Changes:                                               */
/*     *CBS_list: list of servers                         */
/*                                                        */
/* Returns:                                               */
/*      0: Sucess                                         */
/*     -1: Too many servers                               */
/*     -2: Invalid parameters                             */
/*     -3: Message already has a CBS server               */
/*     -4: Unfeasible (not enough bandwidth available)    */
/**********************************************************/
{
  WORD ns;
  int i; 
  double Usm, Ulub, Ucbs;
  
  ns=CBS_list->ns;

  /* Check if space for more servers */
  if(ns == CBS_MAXSERV)
    return -1;

  /* Check input parameters */
  if(Q == 0 || T == 0 || id == 0)
    return -2;

  /*  Check if message already has a server */
  id |= ADATA_MESG_ID_DATA_LONG;
  for(i=0;i<ns;i++)
    if(CBS_list->cbs[i].id == id)
      return -3;

  /* Admission test */
  /*   Note: for use with the Elastic Task Model, a compression can be required - FUTURE WORK */

  /* Get utilization of synchronous messages */
  SRDB_SRT_GetUf(&SRDB_SRT_mesgtable,&Usm,&Ulub);

  cprintf("\n CBS_addserver: Usm=%f Ulub=%f",Usm,Ulub);

  /* Get utilization of current CBSs ... */
  CBS_getUcbs(CBS_list,&Ucbs);

  cprintf(" Ucbs(before/after addition)=%f / ",Ucbs);
  
  /*      ... and of the server being added */
  Ucbs += (double)(Q*8*BIT_TIME_US / (T*EC_LEN) );
  cprintf(" %f", Ucbs);
 
  /* Check if set is feasible */
  if( (Usm+Ucbs) > Ulub) 
    return(-4);

  /* Feasible set : Add server to the list */
  CBS_list->cbs[ns].id = id;
  CBS_list->cbs[ns].Q  = Q;
  CBS_list->cbs[ns].T  = T;
  
  CBS_list->cbs[ns].d = 0;
  CBS_list->cbs[ns].c = Q;
  CBS_list->cbs[ns].n = 0;

  /* Increment the number of servers */
  (CBS_list->ns)++;

  return 0;
}

int CBS_removeserver(CBS_list_type *CBS_list, WORD id )
/****************************************************************/
/* This function removes a CBS server from the CBS servers list */
/*                                                              */
/* Inputs:                                                      */
/*      id: message served by the CBS                           */
/*                                                              */
/* Changes:                                                     */
/*     *CBS_list: list of servers                               */
/*                                                              */
/* Returns:                                                     */
/*      0: Success                                              */
/*     -1: Message not found                                    */
/****************************************************************/
{
  int i,j;
  WORD ns;
  
  ns=CBS_list->ns;

  /* Check if space for more servers */
  if(ns == 0)
    return -1;

  /*  Check if message already has a server */
  id |= ADATA_MESG_ID_DATA_LONG;
  for(i=0;i<ns;i++)
    if(CBS_list->cbs[i].id == id)
      break;

  /* Server not found : exit */
  if(i==ns)
    return -1;  

  /* Server found: remove it from the list */
  for(j=i+1;j<ns;j++)
    CBS_list->cbs[j-1]=CBS_list->cbs[j];

  /* Decrement the number of servers */
  (CBS_list->ns)--;
    
  return 0;
}


void CBS_printservers(CBS_list_type *CBS_list)
/*************************************************************/
/* This function prints the defined CBS servers properties   */
/*  !FOR DEBUG ONLY!                                         */
/*************************************************************/
{
  WORD ns;
  int i,j;
   
  ns=CBS_list->ns;

  cprintf("\n CBS servers list");
  for(i=0;i<ns;i++)
    {
      cprintf("\n[%2d] id=%x Q=%u T=%u (Util=%f)", i, CBS_list->cbs[i].id, \
	      CBS_list->cbs[i].Q, CBS_list->cbs[i].T, \
	      (float)CBS_list->cbs[i].Q*8*BIT_TIME_US / (CBS_list->cbs[i].T*EC_LEN) );
      cprintf("c=%u d=%d n=%u [",CBS_list->cbs[i].c,CBS_list->cbs[i].d,CBS_list->cbs[i].n);
      for(j=0;j<CBS_list->cbs[i].n;j++)
	cprintf(" %u ",CBS_list->cbs[i].dlc[j]);
      cprintf("]");
    }
}



int CBS_request(CBS_list_type *CBS_list, WORD id, WORD dlc)
/**************************************************************************/
/* This function handles a request for transmission made to a CBS server  */
/*                                                                        */
/* Inputs:                                                                */
/*     *CBS_list: list of servers                                         */
/*     id: id of the message                                              */
/*    dlc: number of *data* bytes to transfer                             */
/*                                                                        */
/* Changes:                                                               */
/*     *CBS_list: updates deadlines, budget ...                           */
/*                                                                        */      
/* Returns:                                                               */
/*     0: Sucess                                                          */
/*    -1: No server found for message id                                  */
/**************************************************************************/
{

  WORD ns;
  int i,dlc_ssize;
  long dlc_bits;
  EDF_READY_QUEUE_RECORD new_rec; /* Record in EDF ready message queue */
  
  /* Var init */
  ns=CBS_list->ns;
  id |= ADATA_MESG_ID_DATA_LONG;

  /* Check if any servers defined */
  if(ns == 0)
    return -1;

  /* Find message server */
  for(i=0;i<ns;i++)
    if( CBS_list->cbs[i].id == id)
      break;
  
  if(i == ns)
    return -1; /* Message server not found */


    /*********************************************/
   /* Message server found - update server data */
  /*********************************************/

  /* Each message requires an additional field (the first after the header) */
  /* to carry the data length of the message */
  dlc+=sizeof(WORD);

  /* Compute the real length (data length + overheads) of transaction*/
  if(dlc >= ETH_DATALEN_MIN-sizeof(FTT_ADM_HEADER))
    dlc_bits = dlc * 8 + sizeof(FTT_SDM_HEADER)*8 + ETH_OVERHEAD_bits;
  else
    dlc_bits = ETH_DATALEN_MIN*8 + ETH_OVERHEAD_bits;
  dlc_bits += FTT_IFS_bits;

  /* Bus time is allocated in slots (and thus the number of bits can be further increased */ 
  dlc_ssize = ceil(dlc_bits * BIT_TIME_US / SLOTV_US);
  dlc_bits = dlc_ssize * SLOTV_US / BIT_TIME_US;
  dlc = ceil((double)dlc_bits / 8);

  /* The schedules accesses the EDF ready queue thus mutual exclusion must be guaranteed */ 
  mutex_lock(&srt_access_mutex);

  /* Store request in server */
  CBS_list->cbs[i].dlc[CBS_list->cbs[i].n] = dlc; /* Store request dlc */
  (CBS_list->cbs[i].n) ++; /* Update number of requests */

  /* If server idle update its state and release request */
  if(CBS_list->cbs[i].n == 1)
    { /* CBS server Idle */
     
      cprintf("\nCBS_request: Server Idle");
 
      /* Update server state */
      if(CBS_list->cbs[i].c >= dlc)
	{ /* Server budget enough to serve the request */

	  if(CBS_list->cbs[i].c / CBS_list->cbs[i].Q * CBS_list->cbs[i].T >= CBS_list->cbs[i].d)
	    { /* Rule 1 */
	      cprintf("\nCBS_request: Rule 1 (deadline and budget updated)");
	      CBS_list->cbs[i].d = CBS_list->cbs[i].T; /* Update deadline */
	      CBS_list->cbs[i].c = CBS_list->cbs[i].Q; /* Refill server budget */
	    }
	  else
	    { /* Rule 2 : server deadline unchanged; server budget unchanged*/
	      cprintf("\nCBS_request: Rule 2 (no change sto budget and deadline)");
	    }
	}
      else
	{ /* Server budget not enough to serve the request */
	  cprintf("\nCBS_request: budget not enough for request (budget refilled)");
	  CBS_list->cbs[i].d += CBS_list->cbs[i].T; /* Update deadline */
	  CBS_list->cbs[i].c += CBS_list->cbs[i].Q; /* Refill server budget */
	}

      /* Release the request (put in EDF ready queue) */

      cprintf("\nCBS_request: release CBS message ID=%x deadline=%d Time=%lu", \
	      CBS_list->cbs[i].id,CBS_list->cbs[i].d, sys_gettime(NULL));
      new_rec.m_id       = CBS_list->cbs[i].id;
      new_rec.m_ssize    = dlc_ssize;
      new_rec.m_deadline = CBS_list->cbs[i].d;
      EDF_RQ_insert(&EDF_rq, new_rec);
            
    }

  /* End of mutex code     */ 
  mutex_unlock(&srt_access_mutex);
  
  return 0;

}


#endif /* CBS_SERVER */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/

#ifdef ELASTIC_TM
    /*****************************************************/
   /*****************************************************/
  /*    Functions used by the Elastic Task Model       */
 /*****************************************************/
/*****************************************************/

signed char ETM_compress(SRDB_SRT_tabtype *srt,double Ud)
/***********************************************************************************************/
/*! 
 \brief This function implements the compress/decompression of a message set
 
 This function implements the compress/decompression of a message set passed as an argument.
                                                                                     
\param *srt: pointer to the new message set                                      
                                                                                             
\returns *srt: argument holds the message set with the periods adjusted (if feasible)
\returns return value :\n 
						-ETM_FEASIBLE if a feasible set is found\n                            
                        -ETM_INFEASIBLE if no feasible set is found
						
***********************************************************************************************/
{
  double U_nom, /* Message set utilization factor corresponding to the nominal period */
         U_min, /* Message set utilization factor corresponding to the maximum periods */
    U0v, /* nominal util factor of messages that can still be compressed */
    Uf, /* Total compression utilisation factor corresp a mesgs that can not be more compressed */
    Ev, /* Sum of the elastic coef of compressible tasks */
    Ui, /* Utilization factor of a particular message */
    Uio,/* Nominal UF for a given message */
    Uact;/* UF of the current set */


  /* To keep the algorithm simpler, all original periods are saved and all the computations are */
  /*    performed over the normal message properties. At the end the new periods are compared   */
  /*    with the original ones,and the properties period and period_shadow are updated accordingly */
  unsigned int ori_period[MAX_VAR];

  double Ci; /* Transaction time of a particular message */
  int i,nvars,Ti;
  
  char opt_ok;

  char comp_complete; /* Flag to signal that the job is done */

  float dUi,dUact;
  float dUd[MAX_VAR],V_Ti[MAX_VAR];

  float min_delta;
  int min_pos;

  int moptim[MAX_VAR];

  int m_updates;

  /* Variable initializations */
  U_nom=0;
  U_min=0;
  Uact=0;
  nvars=srt->sm_nvars;
  Ci=0;

  /* Save original periods*/
  for(i=0;i<nvars;i++)
    ori_period[i]=srt->sm_message[i].sm_period;


  /* Get nominal and minimum utilization factors for the message set */
  for(i=0;i<nvars;i++)
    {
      Ci=(double)srt->sm_message[i].sm_ssize*SLOTV_US;
      U_nom+=Ci/(srt->sm_message[i].sm_period_nom*EC_LEN);
      U_min+=Ci/(srt->sm_message[i].sm_period_max*EC_LEN);
    }
  
  /* Ud argument == 0 means use all available bandwidth */ 
  if(Ud == 0)
    SRDB_SRT_GetUf(srt,&Uact,&Ud); 
  
  //cprintf("\n Inside ETM_compress! *** Ud=%f Uact=%f",Ud,Uact);
  
  /* Test if set is schedulable */
  if(Ud < U_min)
    return(ETM_INFEASIBLE);
  
  /* Reset periods */
  for(i=0;i<nvars;i++)
    srt->sm_message[i].sm_period=srt->sm_message[i].sm_period_nom;
  
#if 0    
  /* *** DEBUG - Print message set */
  {
    int lixo;
      
    /* Prints the current message set */
    cprintf("\n <<ETM compress - Message Table to compress>>");
    for(lixo=0;lixo<srt->nvars;lixo++)
      {
	cprintf("\n Message: ID=%x",srt->sm_message[lixo].vid);
	cprintf(" size=%d",srt->sm_message[lixo].size);
	cprintf(" period:%d Min=%d",srt->sm_message[lixo].sm_period,srt->sm_message[lixo].sm_period_min);
	cprintf(" nom=%d",srt->sm_message[lixo].sm_period_nom);
	cprintf(" max=%d",srt->sm_message[lixo].sm_period_max);
	cprintf(" deadline=%d",srt->sm_message[lixo].deadline);
      }
  } /* *** */
#endif 

  
  /* Compression algorithm */
  do
    {
      Uf=0;
      Ev=0;
      U0v=0;
      
      /* Calculate Uf (Util fact of tasks that cant be compressed) and      */
      /*           Ev (sum of elastic coef of tasks that can be compressed) */
      for(i=0;i<nvars;i++)
	if( (srt->sm_message[i].sm_elastic_coef == 0) || (srt->sm_message[i].sm_period==srt->sm_message[i].sm_period_max) )
	  {
	    Ci=(double)srt->sm_message[i].sm_ssize*SLOTV_US;
	    Uf+=Ci/(srt->sm_message[i].sm_period*EC_LEN);
	  }
	else
	  {
	    Ev+=srt->sm_message[i].sm_elastic_coef;
	    Ci=(double)srt->sm_message[i].sm_ssize*SLOTV_US;
	    U0v+=Ci/(srt->sm_message[i].sm_period_nom*EC_LEN);
	  }
      //cprintf("\n<<<< Instancia:Ud=%2.1f%% Uf=%2.1f%% U0v=%2.1f%% Ev=%f >>>>>>",Ud*100,Uf*100,U0v*100,Ev);
      
      /* Scan the message set and calculate the new periods */
      /* Note that there are "length" contraints, therefore during the application of the algorithm */
      /*  some messages can exceed the maximum priod (minimum utilization) and therefore the */
      /*  process must be repeated, after including that specific message in the set of the ones */
      /*  that does not allow any further compression */ 
      
      comp_complete = 1;
      for(i=0;i<nvars;i++)
	{
	  if( (srt->sm_message[i].sm_elastic_coef > 0) && (srt->sm_message[i].sm_period < srt->sm_message[i].sm_period_max))
	    { /* For each message that can be compressed */
	      
	      Ci  = (double)srt->sm_message[i].sm_ssize*SLOTV_US;
	      Uio = Ci/(srt->sm_message[i].sm_period_nom*EC_LEN);
	      Ui  = Uio - (U0v - Ud + Uf)*((double)srt->sm_message[i].sm_elastic_coef/Ev);
	      
	      Ti=ceil((Ci/Ui)/EC_LEN);
	      srt->sm_message[i].sm_period = Ti;
	      
	      //cprintf("\n    > ID:%x (Ti)=Ci/Ui=%6.0fus Ti=%d Ui=%f ei=%f ",srt->sm_message[i].sm_id,(Ci/Ui),Ti,Ui,srt->sm_message[i].sm_elastic_coef);

	      if(Ti > srt->sm_message[i].sm_period_max )
		{
		  srt->sm_message[i].sm_period=srt->sm_message[i].sm_period_max;
		  comp_complete = 0;
		  //cprintf("!! Compress limit reached");
		}
	      
	    }
	}
      
    }while(comp_complete == 0);


#define ETM_OPTIMIZE
#ifdef ETM_OPTIMIZE
      
  //cprintf("\n ETM_compress Stage 2: optimization process");

  /* Optimization */
  Uf=0;
  Ev=0;
  U0v=0;
      
  /* Calculate Uf (Util fact of tasks that cant be compressed ei=0) and */
  /*           Ev (sum of elastic coef of tasks that can be compressed) */
  for(i=0;i<nvars;i++)
    if( srt->sm_message[i].sm_elastic_coef == 0 )
      {
	Ci=(double)srt->sm_message[i].sm_ssize*SLOTV_US;
	Uf+=Ci/(srt->sm_message[i].sm_period*EC_LEN);
      }
    else
      {
	Ev+=srt->sm_message[i].sm_elastic_coef;
	Ci=(double)srt->sm_message[i].sm_ssize*SLOTV_US;
	U0v+=Ci/(srt->sm_message[i].sm_period_nom*EC_LEN);
      }
  
  do
    {
      m_updates = 0;
      /* Mark all messages as NOT optimized */
      for(i=0;i<nvars;i++)
	moptim[i]=0;
  
      /* Calculate Ti (float) for each message */
      for(i=0;i<nvars;i++)
	if( srt->sm_message[i].sm_elastic_coef != 0 && moptim[i] == 0 )
	  { /* For all tasks that can be compressed and not yet optimized*/
	    Ci  = (double)srt->sm_message[i].sm_ssize*SLOTV_US;
	    Uio = Ci/(srt->sm_message[i].sm_period_nom*EC_LEN);
	    Ui  = Uio - (U0v - Ud + Uf)*((double)srt->sm_message[i].sm_elastic_coef/Ev);
	    
	    V_Ti[i] = (Ci/Ui)/EC_LEN;
	    
	  }
  
      //cprintf("\n dUd[]=");
      /* Calculate the (next) Ud at which messages decrease its period */
      for(i=0;i<nvars;i++)
	if( srt->sm_message[i].sm_elastic_coef != 0 && moptim[i] == 0 && srt->sm_message[i].sm_period > V_Ti[i])
	  {
	    Ci  = (double)srt->sm_message[i].sm_ssize*SLOTV_US;
	    dUi=Ci/((srt->sm_message[i].sm_period-1)*EC_LEN) - Ci/(V_Ti[i]*EC_LEN);
	    dUd[i]=dUi*Ev/((double)srt->sm_message[i].sm_elastic_coef);
	    //cprintf("[%d]=%f",i,dUd[i]);
	  }

      /* Get current Util factor */
      Uact=0;
      for(i=0;i<nvars;i++)
	{
	  Ci=(double)srt->sm_message[i].sm_ssize*SLOTV_US;
	  Uact+=Ci/(srt->sm_message[i].sm_period*EC_LEN);
	}
      
      /* Get the dUd in ascending order and change the respective period (if possible) */
      do
	{
	  opt_ok = 1;
	  
	  /* get minor (valid) dUd */
	  min_delta = -1;
	  min_pos = -1;
	  for(i=0;i<nvars;i++)
	    if( srt->sm_message[i].sm_elastic_coef != 0 && moptim[i] == 0 && srt->sm_message[i].sm_period > V_Ti[i])
	      {
		//cprintf("#");
		if(min_delta < 0 || (min_delta >0 && dUd[i] < min_delta))
		  {
		    //cprintf("+");
		    min_delta = dUd[i];
		    min_pos = i;
		  }
	      }
	  
	  /* If found a valid dUd update period (if feasible)*/
	  if(min_delta > 0)
	    {
	      //cprintf("->minor dUd pos=%d",min_pos);
	      i=min_pos;
	      Ci=(double)srt->sm_message[i].sm_ssize*SLOTV_US;
	      dUact = Ci/((srt->sm_message[i].sm_period-1)*EC_LEN) - Ci/((srt->sm_message[i].sm_period)*EC_LEN);
	      if(dUact+Uact<Ud)
		{ /* If resulting set feasible */
		  srt->sm_message[i].sm_period=srt->sm_message[i].sm_period-1; /* Update period */
		  Uact+=dUact; /* Update Current Util fact */
		  moptim[i]=1; /* Mark message as optimized */
		  opt_ok = 0;
		  m_updates++;
		  //cprintf("Update!");
		}
	      else
		{
		  //cprintf("Not feasible");
		  moptim[i]=1; /* Mark message as optimized */
		  opt_ok = 0;
		}
	  
	    }
	  else
	    {
	      //cprintf("No valid dUd!");
	    }

  
	}while(opt_ok == 0);

    }while(m_updates > 0); 

#endif


  /* Update period and period_shadow*/
  for(i=0;i<nvars;i++)
    if(ori_period[i] <= srt->sm_message[i].sm_period)
      { /* Period enlarged-> Safe to update it immediatly */
	srt->sm_message[i].sm_period_shadow = srt->sm_message[i].sm_period;
	srt->sm_message[i].sm_deadline = srt->sm_message[i].sm_period;
      }
    else
      { /* Period reduced: restore initial period and update period_shadow */
	/* New (shorter) period will take efect after next message release */
	/*   Note: ftt_scheduler checks period_shadow in all task releases */
	srt->sm_message[i].sm_period_shadow = srt->sm_message[i].sm_period;
	srt->sm_message[i].sm_period = ori_period[i];
      }

#if 0
  Uact=0;
  for(i=0;i<nvars;i++)
    {
      Ci=(double)srt->sm_message[i].sm_ssize*SLOTV_US;
      Uact+=Ci/(srt->sm_message[i].sm_period*EC_LEN);
    }
  //cprintf("\nResulting Ud=%f Uf=%f ",Ud,Uact);
  
  {
    /* DEBUG - Print message set */
    int lixo;
    /* Prints the current message set */
    cprintf("\n <<ETM_compress - Message Table after ETM_compress>>");
    for(lixo=0;lixo<SRDB_SRT_mesgtable_tmp.nvars;lixo++)
      {
	cprintf("\n Message: ID=%x",srt->sm_message[lixo].vid);
	cprintf(" size=%d",srt->sm_message[lixo].size);
	cprintf(" period:%d Min=%d",srt->sm_message[lixo].sm_period,srt->sm_message[lixo].sm_period_min);
	cprintf(" nom=%d",srt->sm_message[lixo].sm_period_nom);
	cprintf(" max=%d",srt->sm_message[lixo].sm_period_max);
	cprintf(" deadline=%d",srt->sm_message[lixo].deadline);
      }
  }
#endif
  
  return(ETM_FEASIBLE);
}

signed char ETM_changeQOS(SRDB_SRT_mesgtype *aux_var)
/****************************************************************************/
/*!
\brief Changes the QOS of an existing message and updates the Master variables' table

   Changes the QOS of an existing message and updates the Master variables' table
   
 \param aux_var: pointer to struct with properties of variable to change\n
         vid (message ID)\n
         size (data size in bytes)\n
         period,deadline,init (in number of ECs)\n
         elastic coeficient\n
 \param SRDB_SRT_tabtype global variables' table        
 \param SRDB_SRT_mesgtable global variables' table 
     
\returns error code\n
					0 : OK\n
                   -1 : Message not found\n
                   -2 : No feasible set is found\n
                   -3 : invalid properties

*****************************************************************************/
{
  
  int ETM_status;            /* To store the result of the compression operation */
  int msgind;

  //cprintf("\n**** Inside ETM_changeQOS");
  
   /* Check if table is empty. If yes terminate with error */
  if(SRDB_SRT_mesgtable.sm_nvars==0)
    return(-1);

  /* Add header for Synch data message */
  aux_var->sm_id |= SDATA_MESG_ID;

  /* Scan table for the message */
  for(msgind=0;msgind<SRDB_SRT_mesgtable.sm_nvars;msgind++)
    if(SRDB_SRT_mesgtable.sm_message[msgind].sm_id == aux_var->sm_id)
      break;
  
  if(msgind == SRDB_SRT_mesgtable.sm_nvars) /* If message not found exit with error */
    return(-1); /* Note: msgind has the index for the variable */

  /* Check mesg properties */
  if(aux_var->sm_period <= 0) return(-3);
  
  if(aux_var->sm_period_min <= 0) return(-3);
  if(aux_var->sm_period_max < aux_var->sm_period_min) return(-3);
  if(aux_var->sm_period < aux_var->sm_period_min) return(-3);
  if(aux_var->sm_period > aux_var->sm_period_max) return(-3);

  aux_var->sm_period_nom=aux_var->sm_period;

  if(aux_var->sm_elastic_coef < 0) return(-3);

  /* Mark message as *NOT DELETED* */
  aux_var->sm_deleted = 0;
  
  /* Complete the input structure, fields: sm_bsize,sm_ssize  */
  /* Note that Ethernet frames have a minimum length      */
  if(aux_var->sm_size >= (ETH_DATALEN_MIN - sizeof(FTT_SDM_HEADER)))
    aux_var->sm_bsize = aux_var->sm_size*8 + sizeof(FTT_SDM_HEADER)*8 + ETH_OVERHEAD_bits;
  else
    aux_var->sm_bsize= ETH_DATALEN_MIN*8 + ETH_OVERHEAD_bits;
  
  /* Due to the variable prop. time on the bus from station to station, and as well */
  /* software and hardware(NIC) induced jitter, an "jitter" factor must be append   */
  /* to all frames, that must be greater than (Network pro. time + Software jitter) */
  aux_var->sm_bsize += FTT_IFS_bits;

  aux_var->sm_ssize=ceil(aux_var->sm_bsize*BIT_TIME_US/SLOTV_US);
  
  //cprintf("\nETM_changeQOS: ID:%x size(bytes):%u size(us-inc.overheads-slots):%lu period:%u",aux_var->vid,aux_var->sm_size,(long unsigned)(aux_var->sm_ssize*SLOTV_US),aux_var->sm_period);

  /* Elastic Task Model operations         */
  /*    1 - Get a copy of the SRT (SRT_t)  */
  /*    2 - Update message prop on SRT_t   */
  /*    3 - Call ETM_compress              */
  /*    4 - If ETM_compress returns : infeasible;                                    */
  /*                                      return -1                                  */
  /*                                : feasible;                                      */
  /*                                      wait for scheduler to stop (mutex)         */
  /*                                      update SRT with new values                 */
  /*                                      update mutex to allow the scheduler to run */
  /*                                      return OK                                  */

  /* Enter mutex region - access must be serialized in relation to other routines using the SRT */
  mutex_lock(&srt_access_mutex);
  
  /* 1 - Get a copy of the current SRT */
  SRDB_SRT_mesgtable_tmp=SRDB_SRT_mesgtable;

  /* 2 - Update mesg prop on the SRT_t */
  SRDB_SRT_mesgtable_tmp.sm_message[msgind]=*aux_var;
  
  /* 3 - Call ETM_compress */
  ETM_status = ETM_compress(&SRDB_SRT_mesgtable_tmp,0);
  
  /* 4 - Manage the SRT according to the ETM_compress result */
  if(ETM_status == ETM_FEASIBLE)
  { /* New SRT is schedulable -> Update master SRT*/
    
    /* Inside mutex region - Update Master SRT */
    for(msgind=0;msgind<SRDB_SRT_mesgtable.sm_nvars;msgind++)
      { /* For  each mesg in the old SRT */
	
	SRDB_SRT_mesgtable.sm_message[msgind].sm_period_min =SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period_min;
	SRDB_SRT_mesgtable.sm_message[msgind].sm_period_nom =SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period_nom;
	SRDB_SRT_mesgtable.sm_message[msgind].sm_period_max =SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period_max;
	SRDB_SRT_mesgtable.sm_message[msgind].sm_elastic_coef =SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_elastic_coef;
	
	SRDB_SRT_mesgtable.sm_message[msgind].sm_size =SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_size;
	SRDB_SRT_mesgtable.sm_message[msgind].sm_bsize =SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_bsize;
	SRDB_SRT_mesgtable.sm_message[msgind].sm_ssize =SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_ssize;
	
	SRDB_SRT_mesgtable.sm_message[msgind].sm_deleted =SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_deleted;
	
	if(SRDB_SRT_mesgtable.sm_message[msgind].sm_period < SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period)
	  { /* If periods are enlarged, update both period and period_shadow */
	    /*     so changes take effect on next release                    */ 
	    SRDB_SRT_mesgtable.sm_message[msgind].sm_period = SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period;
	    SRDB_SRT_mesgtable.sm_message[msgind].sm_period_shadow = SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period;
	    //cprintf("\n mesg %d period=%d period_shadow=%d",msgind,SRDB_SRT_mesgtable.sm_message[msgind].sm_period,SRDB_SRT_mesgtable.sm_message[msgind].sm_period_shadow);
	  }
	else
	  {
	    /* If periods are reduced, update only period_shadow             */
	    /*     so changes take effect only after next release            */ 
	    SRDB_SRT_mesgtable.sm_message[msgind].sm_period_shadow = SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period;
	    //cprintf("\n mesg %d period_shadow=%d",msgind,SRDB_SRT_mesgtable.sm_message[msgind].sm_period_shadow);
	  }
	
#ifdef RM_SCHED
	/* The RM scheduler works over a message table sorted by increasing periods (decreasing priorities) */
	SRDB_SRT_sorttable();
#endif
      }
    
    /* SRT updated -> exit mutex region and return */
    mutex_unlock(&srt_access_mutex);
    return(0);      
  }
  else
  { /* New SRT is NOT schedulable -> Return error code */
    
    //cprintf("\n *** Set not schedulable *****");
    //keyb_getchar();
    
    /* SRT updated -> exit mutex region and return */
    mutex_unlock(&srt_access_mutex);
    
    return(-2);
  }
  
}

#endif 

  


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/  



    /*****************************************************/
   /*****************************************************/
  /*    SRDB handling functions                        */
 /*****************************************************/
/*****************************************************/

int SRDB_SRT_GetUf(SRDB_SRT_tabtype *srt,double *Uact,double *Ulub)
/***********************************************************************************************/
/*! 				   
	\brief This function returns the utilization factor										   
																							   
	This function returns the utilization factor corresponding to the current  message set\n
    (Synchronous messages) passed as an argument, as well as the limit above which the\n
    set is not guaranteedly schedulable.                                                     

	\param *srt: pointer to the message set  
	\param Uact: current utilization factor                                                    
	\param Ulub: limit below which the set is guaranted schedulable                                                                                         
 																							 
***********************************************************************************************/
{
  int i,nvars,long_mesg;

#ifdef RM_SCHED
  double RM_lub;
#endif
  
  /* Variable initializations */
  *Uact=0;
  *Ulub=0;
  long_mesg=0;
  nvars=srt->sm_nvars;
  
  /* Get utilization factor for the message set */
#ifdef ELASTIC_TM
  /* Note: when using the ETM period_shadow holds the (sometimes future) requirements of the mesage set */ 
  for(i=0;i<nvars;i++)
    (*Uact)+=(double)srt->sm_message[i].sm_ssize*SLOTV_US/(srt->sm_message[i].sm_period_shadow*EC_LEN);
#else
  for(i=0;i<nvars;i++)
    (*Uact)+=(double)srt->sm_message[i].sm_ssize*SLOTV_US/(srt->sm_message[i].sm_period*EC_LEN);
#endif
  
  /* Obtain Ulub */
  /*  Since there is no preemption and a limited time slot, there can be idle-time  */
  /*  The idle-time must be used in the test, and in worst case equals the greatest */
  /*  message on the set - <Simplest test. This can be improved>                    */
  for(i=0;i<nvars;i++)
    if(srt->sm_message[i].sm_ssize > long_mesg) 
      long_mesg=srt->sm_message[i].sm_ssize;
  long_mesg*=SLOTV_US;

#ifdef EDF_SCHED
  /* Note that only SWIN_SHARE(%) of the EC is available to the synch. window */
  *Ulub=SWIN_SHARE-(double)long_mesg/EC_LEN; /* Synch window share minus worst-case inserted idle-time */
#endif
      
#ifdef RM_SCHED
  /* Note that only SWIN_SHARE(%) of the EC is available to the synch. window */
  /* For RM the least upper bound lower than 1 (Liu*Layland n*(2^1/n...       */
  RM_lub = 0.69; /* Least upper bound for nvars=infinit.<Put L&L equation?>   */ 
  *Ulub=SWIN_SHARE*RM_lub-(double)long_mesg/EC_LEN;
#endif
   
  return(0); 
}


int SRDB_SRT_sorttable()
/****************************************************************************/
/*! 											
  \brief This function sorts the SRT message table.						

 This function sorts the SRT message table.\n
 The RM scheduler requires the SRT to be sorted by increasing order of periods\n
       (i.e. decreasing order of priorities)

  \param SRT table (SRDB_SRT_mesgtable) global var
  \returns 0: OK 
  \returns changes SRT table (SRDB_SRT_mesgtable) global var
    
*****************************************************************************/
{


  /****************/
  /* Declarations */
  /****************/

  int mesg_ind;    /* Message index */
  int change_flag;

  SRDB_SRT_mesgtype tmpmesg_s;
  
  /***********************************/
  /*   Code                          */
  /***********************************/
  
  do
    {
      change_flag = 0;
      
      /* For each message in SRT table ... */
      for(mesg_ind = 0 ; mesg_ind < SRDB_SRT_mesgtable.sm_nvars-1; mesg_ind++) 
	{
	  if(SRDB_SRT_mesgtable.sm_message[mesg_ind].sm_period > SRDB_SRT_mesgtable.sm_message[mesg_ind+1].sm_period )
	    {
	      tmpmesg_s = SRDB_SRT_mesgtable.sm_message[mesg_ind];
	      SRDB_SRT_mesgtable.sm_message[mesg_ind] = SRDB_SRT_mesgtable.sm_message[mesg_ind+1];
	      SRDB_SRT_mesgtable.sm_message[mesg_ind+1] = tmpmesg_s;

	      change_flag = 1;	/* Signal that a change was made */
	    }
	}
    }while(change_flag != 0);

  return 0;
}



signed char SRDB_SRT_addmesg(SRDB_SRT_mesgtype *aux_var)
/****************************************************************************/
/*! 
 \brief Appends a message to the master variables' Synchronous Requirements Table

   Appends a message to the master variables' Synchronous Requirements Table\n
   Fills in auxiliar fields: (sm_bsize,sm_ssize ...)

 \param *aux_var: pointer to struct with properties of variable to add:\n
					vid (message ID)\n
					size (data size in bytes)\n
					period,deadline,init (in number of ECs)\n
 \param SRDB_SRT_tabtype global variables' table        
 \param SRDB_SRT_mesgtable global variables' table 
 \returns error code\n  
					0 : OK\n
                   -1 : Table full (v_table->nvars = MAX_VAR)\n
                   -2 : No feasible set is found (if using Elastic Task Model)\n
                   -3 : invalid properties\n
                   -4 : invalid ID (duplicated)\n

*****************************************************************************/
{
  
  int i;
#ifdef ELASTIC_TM
  int ETM_status;            /* To store the result of the compression operation */
  int msgind;
#endif
  

   /* Check if table is full, if yes - terminate with error */
  if(SRDB_SRT_mesgtable.sm_nvars==MAX_VAR)
    return(-1);

  /* Check mesg properties */
  if(aux_var->sm_period <= 0) return(-3);
  if(aux_var->sm_init < 0) return(-3);

#ifdef ELASTIC_TM
  if(aux_var->sm_period_min <= 0) return(-3);
  if(aux_var->sm_period_max < aux_var->sm_period_min) return(-3);
  if(aux_var->sm_period < aux_var->sm_period_min) return(-3);
  if(aux_var->sm_period > aux_var->sm_period_max) return(-3);
  
  if(aux_var->sm_elastic_coef < 0) return(-3);
#endif

  /* Mark message as *NOT DELETED* */
  aux_var->sm_deleted = 0;
  
  /* Add header for Synch data message */
  aux_var->sm_id |= SDATA_MESG_ID;

  /* Check for duplicated ID */
  for(i=0;i<SRDB_SRT_mesgtable.sm_nvars;i++)
    if(SRDB_SRT_mesgtable.sm_message[i].sm_id == aux_var->sm_id)
     return -4;

  /* Complete the input structure, fields: sm_bsize,sm_ssize  */
  /* Note that Ethernet frames have a minimum length      */
  if(aux_var->sm_size >= ETH_DATALEN_MIN-sizeof(FTT_SDM_HEADER))
    aux_var->sm_bsize = aux_var->sm_size*8 + sizeof(FTT_SDM_HEADER)*8 + ETH_OVERHEAD_bits;
  else
    aux_var->sm_bsize= ETH_DATALEN_MIN*8 + ETH_OVERHEAD_bits;
  
  /* Due to the variable prop. time on the bus from station to station, and as well */
  /* software and hardware(NIC) induced jitter, an "jitter" factor must be append   */
  /* to all frames, that must be greater than (Network pro. time + Software jitter) */
  aux_var->sm_bsize += FTT_IFS_bits;

  aux_var->sm_ssize=ceil(aux_var->sm_bsize*BIT_TIME_US/SLOTV_US);
    
#ifdef ELASTIC_TM
  /* If Elastic Task Model is to be used   */
  /*    1 - Get a copy of the SRT (SRT_t)  */
  /*    2 - Add new message to SRT_t       */
  /*    3 - Call ETM_compress if necessary */
  /*    4 - If ETM_compress returns : infeasible, return -1                                */
  /*                                : feasible, wait for scheduler to stop (mutex)         */
  /*                                            update SRT with new values                 */
  /*                                            update mutex to allow the scheduler to run */
  /*                                            return OK                                  */
  
  /* 1 - Get a copy of the current SRT */
  SRDB_SRT_mesgtable_tmp=SRDB_SRT_mesgtable;

  /* 2 - Update mesg field and add it to the SRT_t */
  aux_var->sm_period_nom=aux_var->sm_period;

  SRDB_SRT_mesgtable_tmp.sm_message[SRDB_SRT_mesgtable_tmp.sm_nvars]=*aux_var;
  SRDB_SRT_mesgtable_tmp.sm_nvars++;

  /* 3 - Call ETM_compress*/
  ETM_status=ETM_compress(&SRDB_SRT_mesgtable_tmp,0);
    
  /* 4 - Manage the SRT according to the ETM_compress result */
  if(ETM_status == ETM_FEASIBLE)
    { /* New SRT is schedulable -> Update master SRT*/
      
      /* Enter mutex region - access must be serialized in relation to other routines using the SRT */
      mutex_lock(&srt_access_mutex);
      
      /* Inside mutex region - Update Master SRT */
      for(msgind=0;msgind<SRDB_SRT_mesgtable.sm_nvars;msgind++)
	{ /* For each mesg in the old SRT */
	  
	  if(SRDB_SRT_mesgtable.sm_message[msgind].sm_period < SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period)
	    { /* If periods are enlarged, update both period and period_shadow */
	      /*     so changes take effect on next release                    */ 
	      SRDB_SRT_mesgtable.sm_message[msgind].sm_period = SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period;
	      SRDB_SRT_mesgtable.sm_message[msgind].sm_period_shadow = SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period;
	      SRDB_SRT_mesgtable.sm_message[msgind].sm_deadline = SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period;
	    }
	  else
	    {
	      /* If periods are reduced, update only period_shadow             */
	      /*     so changes take effect only after next release            */ 
	      SRDB_SRT_mesgtable.sm_message[msgind].sm_period_shadow = SRDB_SRT_mesgtable_tmp.sm_message[msgind].sm_period;
	    }
	}
      
      /* Update the period of the message being added as well as               */
      /* setting aux_var. period_shadow = period. This is required because the */
      /*   test and update loop executed before does not update the new var    */
      aux_var->sm_period = SRDB_SRT_mesgtable_tmp.sm_message[SRDB_SRT_mesgtable_tmp.sm_nvars-1].sm_period;
      aux_var->sm_period_shadow=aux_var->sm_period;
      aux_var->sm_deadline = aux_var->sm_period;
   
      /* Add var at the end of the table and update number of vars */
      SRDB_SRT_mesgtable.sm_message[SRDB_SRT_mesgtable.sm_nvars]= *aux_var;
      SRDB_SRT_mesgtable.sm_nvars++;

#ifdef RM_SCHED
      /* The RM scheduler works over a message table sorted by increasing periods (decreasing priorities) */
      SRDB_SRT_sorttable();
#endif
	  
      /* SRT updated -> exit mutex region and return */
      mutex_unlock(&srt_access_mutex);
      return(0);      
    }
  else
    { /* New SRT is NOT schedulable -> Return error code */
      
      //cprintf("\n *** Set not schedulable *****");
      //keyb_getchar();
      
      return(-2);
    }
    

#else /* If not using Elastic task model ... */

  /* Enter mutex region - access must be serialized in relation to the scheduler */
  mutex_lock(&srt_access_mutex);
  
  /* Add var at the end of the table and adjust the total number of vars */
  SRDB_SRT_mesgtable.sm_message[SRDB_SRT_mesgtable.sm_nvars]= *aux_var;
  SRDB_SRT_mesgtable.sm_nvars++;

#ifdef RM_SCHED
      /* The RM scheduler works over a message table sorted by increasing periods (decreasing priorities) */
      SRDB_SRT_sorttable();
#endif
  
  /* SRT updated -> exit mutex region and return */
  mutex_unlock(&srt_access_mutex);

#ifdef ETM_MEASURE
  {
    TIME curtime;
    
    curtime=sys_gettime(NULL);
    evt_act[etoact].inst=curtime;
    evt_act[etoact].id=aux_var->vid;
    evt_act[etoact].action=2;
    etoact++;
    
  }
#endif
        
  return(0);
#endif

}


signed char SRDB_SRT_delmesg(WORD vid)
/****************************************************************************/
/*! 
\brief Deletes a synchronous message from the masters Synchronous Requirements Table

   Deletes a synchronous message from the masters Synchronous Requirements Table

\param vid: message ID to be removed
\param SRDB_SRT_tabtype global variables' table        
\param SRDB_SRT_mesgtable global variables' table      

\return error code\n
					0 : OK\n
                   -1 : Table empty\n
                   -2 : Invalid ID

*****************************************************************************/
{
  
  int msgind;

   /* Check if table is empty. If yes terminate with error */
  if(SRDB_SRT_mesgtable.sm_nvars==0)
    return(-1);

  /* Add header for Synch data message */
  vid |= SDATA_MESG_ID;

  /* Scan table for the message */
  for(msgind=0;msgind<SRDB_SRT_mesgtable.sm_nvars;msgind++)
    if(SRDB_SRT_mesgtable.sm_message[msgind].sm_id == vid)
      break;
  
  if(msgind == SRDB_SRT_mesgtable.sm_nvars) /* If message not found exit with error */
    return(-2);

  /* Mark message as deleted */
  /* Note that the message is kept in the SRT until the instant of the next activation.        */
  /*  This is required because only there its influence in terms of bus utilisation ends       */
  /*  i.e., if it was removed imediatly, another new message could be added, but no timeliness */
  /*  guarantees could be given in this case. Efective removal of the message takes place at   */ 
  /*  the scheduler task, after the message becoming ready again. */
  SRDB_SRT_mesgtable.sm_message[msgind].sm_deleted = 1;

#ifdef ETM_MEASURE
  {
    TIME curtime;
    
    curtime=sys_gettime(NULL);
    evt_act[etoact].inst=curtime;
    evt_act[etoact].id=vid;
    evt_act[etoact].action=1;
    etoact++;
    
  }
#endif

  return(0);
  
}

signed char SRDB_SRT_printmesg()
/****************************************************************************/
/*! 
\brief Prints the Synchronous Requirements Table 

Prints the Synchronous Requirements Table (For debug!)
\param SRDB_SRT_tabtype global variables' table        
\param SRDB_SRT_mesgtable global variables' table      

\return error code\n
					0 : OK\n
                   -1 : Table empty

*****************************************************************************/
{
  
  int i;

   /* Check if table is empty. If yes terminate with error */
  if(SRDB_SRT_mesgtable.sm_nvars==0)
    return(-1);

  /* Scan and print table */
  cprintf("\n SRDB_SRT table contents (%u mesgs stored)",SRDB_SRT_mesgtable.sm_nvars);
  for(i = 0 ; i < SRDB_SRT_mesgtable.sm_nvars ; i++)
    {
      cprintf("\n[%2d] ID:%4x Size:%4d/%4d(B/b)", i , SRDB_SRT_mesgtable.sm_message[i].sm_id , SRDB_SRT_mesgtable.sm_message[i].sm_size,(int)SRDB_SRT_mesgtable.sm_message[i].sm_bsize);
      cprintf(" P:%2u D:%2u",SRDB_SRT_mesgtable.sm_message[i].sm_period, SRDB_SRT_mesgtable.sm_message[i].sm_deadline);
#ifdef ELASTIC_TM
      cprintf("[ETM] P(Min:%2u Max:%2u Sh:%2u) Ei:%2.1f",SRDB_SRT_mesgtable.sm_message[i].sm_period_min,SRDB_SRT_mesgtable.sm_message[i].sm_period_max,SRDB_SRT_mesgtable.sm_message[i].sm_period_shadow,SRDB_SRT_mesgtable.sm_message[i].sm_elastic_coef);
#endif
    }  
  
  return 0;
}


signed char SRDB_ART_addmesg(SRDB_ART_mesgtype *aux_var, unsigned int msg_type, unsigned char timeliness_constr)
/****************************************************************************/
/*! 
\brief Appends a message to the master variables' Asynchronous Requirements Table

   Appends a message to the master variables' Asynchronous Requirements Table\n
   Fills in auxiliar fields: (b_size,s_size ...)

\param *aux_var: pointer to struct with properties of variable to add\n
         vid (message ID)\n
         size (data size in bytes)\n
         period,deadline,init (in number of ECs)\n
\param SRDB_ART_tabtype global variables' table
\param SRDB_ART_mesgtable global variables' table
 
\return error code\n
					0 : OK\n
                   -1 : Table full (table->nvars = MAX_VAR)\n
                   -2 : Set unschedulable\n
                   -3 : Invalid message properties\n
                   -4 : Invalid message type\n
                   -5 : Invalid ID (duplicated)

*****************************************************************************/
{

  int i,j; /* General counters */

  //cprintf("\n[SRDB_ART_addmesg] input args: msg_type=%x timeliness_constr=%x",msg_type, timeliness_constr);

  /* Check if table is full, if yes - terminate with error */
  if(SRDB_ART_mesgtable.am_nvars == MAX_VAR)
    return(-1);

  /* Check mesg properties */
  if(aux_var->am_period <= 0) return(-3);
  if(aux_var->am_deadline <= 0) return -3;
  if(aux_var->am_init < 0) return(-3);

  switch(timeliness_constr)
    {
    case TMLN_UNCONST:
      aux_var->am_timeliness = TMLN_UNCONST;
      break;
    case TMLN_HARD:
      aux_var->am_timeliness = TMLN_HARD;
      break;
    case TMLN_SOFT:
      aux_var->am_timeliness = TMLN_SOFT;
      break;

    default:
      return -3;
    }

  //printf("SRDB_ART_addmesg: id:%x msg_type=%x size:%lu ADATA_MESG_ID_DATA_SHORT=%x",aux_var->am_id , aux_var->am_type , aux_var->am_size , (unsigned int)(ADATA_MESG_ID_DATA_SHORT));

  /* Check data size */
  switch(msg_type)
    {
    case ADATA_MESG_ID_DATA_SHORT:
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:
      //printf("\n tipo curto");
      if(aux_var->am_size > (ETH_DATALEN_MIN - sizeof(FTT_ADM_HEADER))) 
	return(-3);
      break;
      
    case ADATA_MESG_ID_DATA_LONG:
      //printf("\n tipo longo");
      if(aux_var->am_size > (ETH_DATALEN_MAX-sizeof(FTT_ADM_HEADER))) 
	return -3;
      break;

    default:
      return -4;
    }
  
  /* Add header for Asynch data message */
  switch(msg_type)
    {
    case ADATA_MESG_ID_DATA_SHORT:
      aux_var->am_id |= ADATA_MESG_ID_DATA_SHORT;
      aux_var->am_type = ADATA_MESG_ID_DATA_SHORT;
      //cprintf("\n[SRDB_ART_addmesg] DATA SHORT");
      break;

    case ADATA_MESG_ID_ALARM:
      aux_var->am_id |= ADATA_MESG_ID_ALARM;
      aux_var->am_type = ADATA_MESG_ID_ALARM;
      //cprintf("\n[SRDB_ART_addmesg] DATA ALARM");
      break;

    case ADATA_MESG_ID_SIGNAL:
      aux_var->am_id |= ADATA_MESG_ID_SIGNAL;
      aux_var->am_type = ADATA_MESG_ID_SIGNAL;
      //cprintf("\n[SRDB_ART_addmesg] DATA SIGNAL");
      break;

    case ADATA_MESG_ID_DATA_LONG:
      aux_var->am_id |= ADATA_MESG_ID_DATA_LONG;
      aux_var->am_type = ADATA_MESG_ID_DATA_LONG;
      //cprintf("\n[SRDB_ART_addmesg] DATA LONG");
      break;

    default:
      return -4;
    }

  /* Check for duplicated ID */
  for(i=0;i<SRDB_ART_mesgtable.am_nvars;i++)
    if(SRDB_ART_mesgtable.am_message[i].am_id == aux_var->am_id)
      return -5;
  
  /* Complete the input structure, fields: b_size,s_size  */
  /* Note that Ethernet frames have a minimum length      */
  if(aux_var->am_size >= (ETH_DATALEN_MIN-sizeof(FTT_ADM_HEADER)))
    aux_var->am_bsize = aux_var->am_size*8 + sizeof(FTT_ADM_HEADER)*8 + ETH_OVERHEAD_bits;
  else
    aux_var->am_bsize = ETH_DATALEN_MIN*8 + ETH_OVERHEAD_bits;
  
  /* Due to the variable prop. time on the bus from station to station, and as well */
  /* software and hardware(NIC) induced jitter, an "jitter" factor must be append   */
  /* to all frames, that must be greater than (Network pro. time + Software jitter) */
  aux_var->am_bsize += FTT_IFS_bits;

  aux_var->am_ssize=ceil(aux_var->am_bsize*BIT_TIME_US/SLOTV_US);

  //cprintf("\n Antes do mutex (1)");
  /* Enter mutex region - access must be serialized in relation to the scheduler */
  mutex_lock(&srt_access_mutex);
  //cprintf("\n Depois do mutex (1)");
  
  /* Add message to the table and adjust the total number of vars */
  /* The table is kept in increasing period order */
  for(i=0;i<SRDB_ART_mesgtable.am_nvars;i++)
    {
      if(SRDB_ART_mesgtable.am_message[i].am_period > aux_var->am_period)
	{ 
	  /* If period greater than the one in position i */
	  /* Insert new message in this point => copy following messages one position ahead */ 
	  for(j = (SRDB_ART_mesgtable.am_nvars-1);j>=i;j--)
	    SRDB_ART_mesgtable.am_message[j+1] = SRDB_ART_mesgtable.am_message[j];
	  
	  break; /* Exit for cycle */
	}
    }
  //printf(" Adicionar em %d",i);
  /* Store message in position i and update number of messages */
  SRDB_ART_mesgtable.am_message[i] = *aux_var;
  SRDB_ART_mesgtable.am_nvars++;

  //cprintf("\n Antes do mutex(2)");
  /* ART updated -> exit mutex region and return */
  mutex_unlock(&srt_access_mutex);
  //cprintf("\n Depois do mutex(2)");
        
  return 0;

}


signed char SRDB_ART_delmesg(WORD vid)
/****************************************************************************/
/*! 
\brief Deletes an asynchronous message from the masters Asynchronous Requirements Table 

   Deletes an asynchronous message from the masters Asynchronous Requirements Table

\param vid: message ID to be removed  
\param SRDB_ART_tabtype global variables' table
\param SRDB_ART_mesgtable global variables' table

\return error code\n
					  0 : OK\n
                     -1 : Table empty\n
                     -2 : Invalid ID

*****************************************************************************/
{
  
  int i,j;

   /* Check if table is empty. If yes terminate with error */
  if(SRDB_ART_mesgtable.am_nvars==0)
    return(-1);

  /* Scan table for the message */
  for(i = 0 ; i < SRDB_ART_mesgtable.am_nvars ; i++)
    if( (SRDB_ART_mesgtable.am_message[i].am_id & MESGID_MASK) == vid)
      break;
  
  /* If message not found exit with error */
  if(i == SRDB_ART_mesgtable.am_nvars) 
    return(-2);

#ifndef teste   
  /* Enter mutex region - access must be serialized in relation to the scheduler */
  mutex_lock(&srt_access_mutex);
#endif
  
  /* Message found. Delete it from the ART */
  for(j = i+1; j < (SRDB_ART_mesgtable.am_nvars-1);j++)
    SRDB_ART_mesgtable.am_message[j] = SRDB_ART_mesgtable.am_message[j+1];
  
  /* Update number of messages */
  SRDB_ART_mesgtable.am_nvars--;

#ifndef teste 
  /* ART updated -> exit mutex region and return */
  mutex_unlock(&srt_access_mutex);
#endif
  
  return(0);
}

signed char SRDB_ART_printmesg()
/****************************************************************************/
/*! 
\brief Prints the Asynchronous Requirements Table

   Prints the Asynchronous Requirements Table (For debug!)

\param SRDB_ART_tabtype global variables' table
\param SRDB_ART_mesgtable global variables' table  

\return error code\n
					0 : OK\n
                   -1 : Table empty

*****************************************************************************/
{
  
  int i;

   /* Check if table is empty. If yes terminate with error */
  if(SRDB_ART_mesgtable.am_nvars==0)
    return(-1);

  /* Scan and print table */
  cprintf("\n SRDB_ART table contents (%u mesgs stored)",SRDB_ART_mesgtable.am_nvars);
  for(i = 0 ; i < SRDB_ART_mesgtable.am_nvars ; i++)
    {
      cprintf("\n[%2d] ID:%4x t:",i, SRDB_ART_mesgtable.am_message[i].am_id);
      switch(SRDB_ART_mesgtable.am_message[i].am_type)
	{
	case  ADATA_MESG_ID_DATA_SHORT:
	  cprintf("D_SHORT");
	  break;
	case ADATA_MESG_ID_ALARM:
	  cprintf("ALARM  ");
	  break;
	case  ADATA_MESG_ID_SIGNAL:
	  cprintf("SIGNAL ");
	  break;
	case  ADATA_MESG_ID_DATA_LONG:
	  cprintf("D_LONG ");
	  break;
	}
      cprintf(" d_siz(Byte):%4d", SRDB_ART_mesgtable.am_message[i].am_size);
      cprintf(" bsiz:%4lu ssiz:%d per:%u",SRDB_ART_mesgtable.am_message[i].am_bsize,SRDB_ART_mesgtable.am_message[i].am_ssize,SRDB_ART_mesgtable.am_message[i].am_period);
      cprintf(" ddln:%u init:%d ",SRDB_ART_mesgtable.am_message[i].am_deadline,SRDB_ART_mesgtable.am_message[i].am_init);
    }  
  
  return 0;
}


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/




#ifdef EDF_SCHED


  /******************************************/
 /* EDF ready queue manipulation functions */
/******************************************/

int EDF_RQ_init(EDF_READY_QUEUE *ready_queue)
/***************************************************/
/*!
\brief Init EDF ready queue data structures

Init EDF ready queue data structures 
***************************************************/
{
  int i;

  ready_queue->num_recs=0;

  for(i=0;i<MAX_READY_MESGS;i++)
    ready_queue->rec[i].r_free=EDF_RQ_FREE;

  return 0;
}

int EDF_RQ_insert(EDF_READY_QUEUE *ready_queue, EDF_READY_QUEUE_RECORD new_rec)
/**************************************************************************************/
/*!
\brief Inserts a new record at the end of the ready queue  

Inserts a new record at the end of the ready queue\n 
 !!Note!! The queue becomes UNSORTED. It must be sorted before use for dispatching
 
 \returns  0: OK\n                                                                  
           -1: Queue Full                                                          
**************************************************************************************/
{

  /* Check if queue not full */
  if(ready_queue->num_recs >= MAX_READY_MESGS)
    return -1;

  /* Copy message data from record to queue (at the last position of the queue) */
  ready_queue->rec[ready_queue->num_recs].m_id = new_rec.m_id;
  ready_queue->rec[ready_queue->num_recs].m_ssize = new_rec.m_ssize;
  ready_queue->rec[ready_queue->num_recs].m_deadline = new_rec.m_deadline;

  /* Set record as used in the ready queue */
  ready_queue->rec[ready_queue->num_recs].r_free = EDF_RQ_BUSY;

  /* Update number of records in use */
  (ready_queue->num_recs)++;
  
  return 0;
}

int EDF_RQ_sort_compar(const void *rec1, const void *rec2)
/***************************************************/
/*!
\brief Compare function used by qsort()

Compare function used by qsort()

 \returns -1: if rec1 < rec 2\n                  
           0: if rec1 == rec2\n                  
           1: if rec1 > rec2
		   
***************************************************/
{
  EDF_READY_QUEUE_RECORD *r1, *r2;

  r1=(EDF_READY_QUEUE_RECORD *)rec1;
  r2=(EDF_READY_QUEUE_RECORD *)rec2;
  
  if( r1->m_deadline < r2->m_deadline)
    return -1;
  else
    if( r1->m_deadline == r2->m_deadline)
      return 0;
    else
      return 1;
}

int EDF_RQ_sort(EDF_READY_QUEUE *ready_queue)
/**************************************************************************************/
/*!
\brief Sorts the EDF ready message queue. 

 Sorts the EDF ready message queue.\n  
 Note: sorting is required before the use of the queue for dispatching, since\n       
       the dispatcher considers at input a sorted queue of ready messages.          
**************************************************************************************/
{
  // !NOTE! qsort() not available in SHARK
  //if(ready_queue->num_recs > 0)
  //  {
  //    /* Use qsort (in stdlib) for sort the EDF ready queue */
  //    /* Usage: qsort( void *base, size_t nmemb, size_t size, int (*compar)(const void *, const void *)) */  
  //    qsort( &(ready_queue->rec[0]), ready_queue->num_recs, sizeof(EDF_READY_QUEUE_RECORD), &EDF_RQ_sort_compar);
  //  }

  int i,j, nr,swap_flag;
  EDF_READY_QUEUE_RECORD rq_rec;

  /* Get number of records */
  nr = ready_queue->num_recs;

  /* If empty or one record no need for sorting */
  if(nr < 2) return 0;

  /* Sort records (ascending order of deadlines) */
  for(i=0 ; i<nr ; i++)
    {
      swap_flag = 0;
      for(j=0 ; j<nr-i-1 ; j++)
	{
	  if(ready_queue->rec[j].m_deadline > ready_queue->rec[j+1].m_deadline)
	    {
	      /* Swap records */
	      rq_rec = ready_queue->rec[j];
	      ready_queue->rec[j] = ready_queue->rec[j+1];
	      ready_queue->rec[j+1] = rq_rec;
	      
	      /* Signal that queue not totally ordered */
	      swap_flag = 1;
	    } 
	}
      if(swap_flag == 0) break;
    }
  
  return 0;
}

int EDF_RQ_update_deadlines(EDF_READY_QUEUE *ready_queue, unsigned int *md_id)
/*****************************************************************************/
/*!
\brief Updates the (absolute) deadlines of ready messages in the ready queue

 Updates the (absolute) deadlines of ready messages in the ready queue.\n
 !NOTE!: this function does NOT modify the sorting of the EDF ready queue.
 
 \returns  0: OK\n                                                      
            -1: missed deadline\n                                         
                 -md_id returns the message ID of first message that missed the deadline                                
*****************************************************************************/
{
  int i;
    
  for(i=0;i<ready_queue->num_recs;i++) /* For each message in queue ... */
    {
      if(ready_queue->rec[i].m_deadline > 0)
	(ready_queue->rec[i].m_deadline)--; /* Update deadline */
      else
	{
	  *md_id = ready_queue->rec[i].m_id; /* Missed deadline. Store ID in return field ... */
	  return -1;  /* and return error */
	}
    }
  return 0;
}

int EDF_RQ_remove_released(EDF_READY_QUEUE *ready_queue, EDF_READY_QUEUE *ready_queue_tmp)
/*****************************************************************************/
/*!
\brief Removes the released messages from the EDF ready queue

 Removes the released messages from the EDF ready queue.\n
   Messages released are signaled by EDF_RQ_FREE in the r_free field.\n        
 !NOTE!: this function does NOT modify the sorting of the EDF ready queue.\n  
   Updates the number of records in queue field.                         
*****************************************************************************/
{
  int i,j=0;
    
  /* Copy used records for tmp queue */
  for(i=0;i<ready_queue->num_recs;i++) /* For each message in queue ... */
    {
      if(ready_queue->rec[i].r_free == EDF_RQ_BUSY )
	{
	  ready_queue_tmp->rec[j] = ready_queue->rec[i];
	  j++;
	}
    }
  ready_queue->num_recs=j;
  
  if(ready_queue->num_recs > 0)
    {
      /* Copy back data */
      memcpy( &(ready_queue->rec[0]), &(ready_queue_tmp->rec[0]), \
	      sizeof(EDF_READY_QUEUE_RECORD) * ready_queue->num_recs );  
    }
  return 0;
}


int EDF_RQ_printqueue(EDF_READY_QUEUE *ready_queue)
/*****************************************************************************/
/*!
\brief Prints the EDF ready queue contents                            

Prints the EDF ready queue contents (for debug).     
*****************************************************************************/
{
  int i;
  
  cprintf("\n EDF ready queue contents:");
  for(i=0;i<ready_queue->num_recs;i++) /* For each message in queue ... */
    {
      
      cprintf("\n ID: %x size:%u deadline=%u free:%d",ready_queue->rec[i].m_id, ready_queue->rec[i].m_ssize, \
	     ready_queue->rec[i].m_deadline, ready_queue->rec[i].r_free);
      
    }
  cprintf("\n");

  return 0;
}

#endif /* EDF_SCHED */

/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/




    /*****************************************************/
   /*****************************************************/
  /*    FTT Tasks implementation                       */
 /*****************************************************/
/*****************************************************/

TASK ftt_scheduler(void *arg)
/****************************************************************************/
/*!
\brief Master scheduler routine

   Master scheduler routine/n
     scans the message tables/n
     builds the EC-schedule/n
     each EC has a list of messages to be produced in it messages\n
      in each EC are identified by its ID.

This task is aperiodic, and is triggered by the dispatcher,\n
	  whenever a EC has been dispatched. The idea is  to have 2 ECs:\n
              while the dispatcher processes one, the scheduler builds the other.


 \param SRDB_SRT_mesgtable: global Master Message Table 
 \param SRDB_ART_mesgtable: global Master Message Table\n
 
 global pointer to EC schedule record (sched_ec) to be built
 
 \returns error code\n
						0 : OK\n
\returns MISSED_DEADLINE - self explanatory!\n
\returns a status code is used within the EC-schedule for catching timing errors

*****************************************************************************/
{


  /****************/
  /* Declarations */
  /****************/

  unsigned int j;          /* General counter */
  unsigned char del_pend;  /* Flag to signal if are any pending request for delete a message */ 

  int mesg_ind;    /* Message index */

  /* Pointer for messages in the SYNC/ASYNC tables */
  SRDB_SRT_mesgtype *curmesg_s=NULL;
  SRDB_ART_mesgtype *curmesg_a=NULL;

#ifdef EDF_SCHED
  EDF_READY_QUEUE_RECORD new_rec; /* Record in EDF ready message queue */
  unsigned int md_id;   /* ID of message that has missed its deadline (if any) */ 
#endif

/**********************************************************************************************/
/* Note: if CBS_SERVER  the following global variables (for exclusive use of the scheduler)   */
/*       are declared:                                                                        */
/*  WORD SCHED_CBS_mlist[]; -> List ofids of messages scheduled for tramsmission in one EC    */
/*  int SCHED_CBS_nm; -> Number of messages scheduled for tramsmission in one EC              */
/**********************************************************************************************/
#ifdef CBS_SERVER
  int ns,i;
  int cbs_ind;
#endif

 /***********************************/
 /*   Code                          */
/***********************************/

  /* Initializations */
  del_pend = 0; /* No pending delete actions */

#ifdef EDF_SCHED
  /* Init EDF ready queue */
  EDF_RQ_init(&EDF_rq);
#endif
  
  while(1)
    {

      /*******************/
      /* Initializations */
      /*******************/
      /* Start to build next EC-schedule */
      sched_ec->status=BUSY_SCHED;	   /* Signal that plan is being built */
      sched_ec->err_code=SCH_ERRCODE_OK;   /* No errors */
      sched_ec->sm_load=0;
      sched_ec->sm_nmesgs=0;
      sched_ec->am_load=0;
      sched_ec->am_nmesgs=0;

      /* Clear the EC-schedule ID lists (sync and async) */
      for(j=0; j<EC_MAX_VAR; j++) 
	{
	  sched_ec->EC_SIDlist[j]=0;
	  sched_ec->EC_AIDlist[j]=0;
	}

#ifdef CBS_SERVER
      /* Init the data structures neede to Store IDs of messages served by the CBS servers */ 
      SCHED_CBS_nm=0;
#endif      


        /*************************************/
       /* Schedule the synchronous messages */
      /*************************************/

      /* SRT must be updated when it is not being used by the scheduler */
      /* All other tasks using the SRT must check this mutex            */ 
      mutex_lock(&srt_access_mutex);

      del_pend = 0; /* Flag to signal instances of deleted messages */
#ifdef RM_SCHED 
    
      /************************************************************/
      /* Build the EC-schedule - Rate Monotonic scheduling policy */
      /************************************************************/
	  
      /* For each message in SRT table ... */
      for(mesg_ind = 0 ; mesg_ind < SRDB_SRT_mesgtable.sm_nvars; mesg_ind++) 
	{
	  /* Scan SRT table. Messages are ordered by decreasing priority, thus to build RM schedules */
	  /*   its enough to scan the table in ascending order */
	  curmesg_s = &(SRDB_SRT_mesgtable.sm_message[mesg_ind]); /* Get copy of current message properties */
	  
	  //cprintf(" Check mesg %x;",curmesg_s->sm_id);
	  
	  if(curmesg_s->sm_deleted == 0 && curmesg_s->sm_init <= 0 && \
	     ( sched_ec->sm_load + curmesg_s->sm_ssize*SLOTV_US ) < EC_MAX_us_SP )
	    { /* Message ready, fits within the Synch window of the EC and was not deleted*/
	      
	      //cprintf("mesg sched;");
	      
	      sched_ec->EC_SIDlist[sched_ec->sm_nmesgs] = curmesg_s->sm_id; /* Store ID */
	      sched_ec->EC_SSizelist[sched_ec->sm_nmesgs] = curmesg_s->sm_ssize;
	      sched_ec->sm_nmesgs++; /* Update number of synch. messages in EC */
	      sched_ec->sm_load += (curmesg_s->sm_ssize*SLOTV_US); /* Update the synch load */
	      
#ifdef ELASTIC_TM
	      /* If the Elastic Task Model is being used, the period_shadow holds          */
	      /*  the new period of the message (after compression). The period            */
	      /*  should be updated only after the next release if there is an decrease    */
	      /*  in it.                                                                   */		  
	      if(curmesg_s->sm_period != curmesg_s->sm_period_shadow)
		{
		  /* If period change, update period and deadline and ... */
		  curmesg_s->sm_period   = curmesg_s->sm_period_shadow;
		  curmesg_s->sm_deadline = curmesg_s->sm_period_shadow;
		  
		  /* sort the resulting message set by decreasing priorities */ 
		  SRDB_SRT_sorttable();
		}
#endif
	      /* Message scheduled. Update the next release instant */
	      curmesg_s->sm_init += curmesg_s->sm_period;
	      
	      //cprintf(" next release=%d.",curmesg_s->sm_init);
	      
	    }
	  else
	    {
	      //cprintf(" mesg !NOT! sched.");
	    }
	  
	  /* Check if a "deleted" messages was released. */ 
	  if(curmesg_s->sm_deleted !=0 && curmesg_s->sm_init <= 0)
	    {
	      curmesg_s->sm_deleted = 2; /* Signal that message can be removed from the SRT */
	      del_pend = 1;
	    }
	  
	} /* For each message in SRT */
      
      
      /***********************/
      /* Update the releases */
      /***********************/
      //cprintf("\n Update releases!");
      for(mesg_ind = 0 ; mesg_ind < SRDB_SRT_mesgtable.sm_nvars; mesg_ind++) /* For each message in SRT table */
	{
	  curmesg_s = &(SRDB_SRT_mesgtable.sm_message[mesg_ind]); /* Get copy of current message properties */
	  
	  //cprintf(" Check mesg %x;",curmesg_s->sm_id);
	  /* Update the time until release */
	  curmesg_s->sm_init--;
	  
	  //cprintf(" updated init=%d (period=%d);",curmesg_s->sm_init,curmesg_s->sm_period);
	  
	  /* Check for missed deadlines */
	  if( (curmesg_s->sm_init <= -(signed int)(curmesg_s->sm_period) ) && (curmesg_s->sm_deleted == 0 ))
	    {
	      sched_ec->status=MISSED_DEADLINE;
	      sched_ec->err_code=SCH_ERRCODE_SM_MISS_DEADLINE;
	      cprintf("\n ! Scheduler fatal error: Missed deadline in Synchronous message (RM ID=%x)",curmesg_s->sm_id);
	      
	    }
	} /* For each message in SRT table ... */
      
#endif /* RM Scheduling */
      

      
#ifdef EDF_SCHED
      /*********************************************************************************/
      /* Very basic and simple alghoritm (to be improved) for EDF message scheduling   */
      /*   1 - Compute activations and insert active messages in the EDF ready queue   */
      /*   2 - After all activations computed, sort the EDF ready queue                */
      /*   3 - Scan the EDF ready queue for schedulable messages (queue is ordered     */
      /*       according to message absolute deadlines)                                */
      /*   4 - Remove scheduled messages from the EDF ready queue                      */
      /*   5 - Update deadlines of messages in the ready queue                         */ 
      /*********************************************************************************/

      /* 1- Compute message activations in current EC */
      for(mesg_ind=0 ; mesg_ind < SRDB_SRT_mesgtable.sm_nvars ; mesg_ind++) /* For each message in the SRT*/
	{  
	  curmesg_s = &(SRDB_SRT_mesgtable.sm_message[mesg_ind]);
	  (curmesg_s->sm_init)--;
	  if(curmesg_s->sm_init < 0 && curmesg_s->sm_deleted == 0)
	    {
	      /* Message ready and not deleted: put in EDF ready queue ... */
	      new_rec.m_id=curmesg_s->sm_id;
	      new_rec.m_ssize=curmesg_s->sm_ssize;
	      new_rec.m_deadline=curmesg_s->sm_deadline;
	      EDF_RQ_insert(&EDF_rq, new_rec);
	      //cprintf(" %x,",curmesg_s->sm_id);
	      
	      /* ... and update sm_init for next release */
	      curmesg_s->sm_init = curmesg_s->sm_period - 1 ;

#ifdef ELASTIC_TM
	      /* If the Elastic Task Model is being used, the period_shadow holds          */
	      /*  the new period of the message (after compression). The period            */
	      /*  should be updated only after the next release if there is an decrease    */
	      /*  in it.                                                                   */		  
	      if(curmesg_s->sm_period != curmesg_s->sm_period_shadow)
		{
		  /* If period change, update period and deadline and ... */
		  curmesg_s->sm_period   = curmesg_s->sm_period_shadow;
		  curmesg_s->sm_deadline = curmesg_s->sm_period_shadow;
		}
#endif
	    }

	  /* Check if a "deleted" messages was released. */ 
	  if(curmesg_s->sm_deleted == 1 && curmesg_s->sm_init < 0)
	    {
	      curmesg_s->sm_deleted = 2; /* Signal that message can be removed from the SRT */
	      del_pend = 1;
	    }
	}
      
      /* 2- Sort EDF ready queue */
      EDF_RQ_sort(&EDF_rq);
      
      /* 3- Build EC schedule */
      for(j=0;j<EDF_rq.num_recs;j++) /* For each message in queue, starting from the lowest abs deadline ... */
	{
	  if( ( (sched_ec->sm_load + EDF_rq.rec[j].m_ssize*SLOTV_US ) < EC_MAX_us_SP ) && EDF_rq.rec[j].r_free == EDF_RQ_BUSY)
	    { /* If Message is ready and fits in the EC ... */
	      
	      /* Update the EC-Schedule and ...*/
	      sched_ec->EC_SIDlist[sched_ec->sm_nmesgs] = EDF_rq.rec[j].m_id;
	      sched_ec->EC_SSizelist[sched_ec->sm_nmesgs] = EDF_rq.rec[j].m_ssize;
	      sched_ec->sm_load += (EDF_rq.rec[j].m_ssize*SLOTV_US); /* Update the synch load */
	      sched_ec->sm_nmesgs++;	
	      
	      /* ... mark the message as removed in the EDF ready queue */
	      EDF_rq.rec[j].r_free = EDF_RQ_FREE;

#ifdef CBS_SERVER
	      /* Store IDs of messages served by the CBS servers */ 
	      if( (EDF_rq.rec[j].m_id &  MESGTYPE_MASK) == ADATA_MESG_ID_DATA_LONG )
		{
		  SCHED_CBS_mlist[SCHED_CBS_nm] = EDF_rq.rec[j].m_id;
		  SCHED_CBS_nm++;
		}
#endif
	    }
	}

      /* 4- Remove scheduled messages from the EDF ready queue */
      EDF_RQ_remove_released(&EDF_rq, &EDF_rqtmp);
      
      /* 5- Update deadlines of ready messages in the EDF ready queue */
      if(EDF_RQ_update_deadlines(&EDF_rq, &md_id))
	{
	  sched_ec->status=MISSED_DEADLINE;
	  sched_ec->err_code=SCH_ERRCODE_SM_MISS_DEADLINE;
	  cprintf("\n ! Scheduler fatal error: Missed deadline in Synchronous message (EDF ID=%x)",md_id);
	}


#ifdef CBS_SERVER
      /* 6 - Handle the CBS servers */
      
      ns=CBS_list.ns; /* ns holds the number of active servers */

      /* Update server deadlines  !!! No inicio ou no fim ???? !!!!*/
      for(cbs_ind=0; cbs_ind<ns; cbs_ind++)
	CBS_list.cbs[cbs_ind].d--;

      /* Remove served messages from the queue of pending requests */
      /*  and update the respective server budget                  */
      for(j=0; j<SCHED_CBS_nm; j++)
	{
	  for(cbs_ind=0; cbs_ind<ns; cbs_ind++)
	    {
	      if( CBS_list.cbs[cbs_ind].id == SCHED_CBS_mlist[j])
		{ /* Served message found */

		  cprintf("\n ftt_scheduler: served message ID=%x Time=%lu",CBS_list.cbs[cbs_ind].id,sys_gettime(NULL));
		  
		  /* Update server budget */ 
		  CBS_list.cbs[cbs_ind].c -= CBS_list.cbs[cbs_ind].dlc[0];

		  /* Remove message from the pending requests queue */
		  for(i=1; i< CBS_list.cbs[cbs_ind].n; i++)
		    CBS_list.cbs[cbs_ind].dlc[i-1] = CBS_list.cbs[cbs_ind].dlc[i];
		  CBS_list.cbs[cbs_ind].n --;

		  /* Check if server budget must be recharged */
		  if(CBS_list.cbs[cbs_ind].c == 0)
		    {
		      cprintf("\n ftt_scheduler: budget zero - refill and update deadline");

		      CBS_list.cbs[cbs_ind].c  = CBS_list.cbs[cbs_ind].Q;
		      CBS_list.cbs[cbs_ind].d += CBS_list.cbs[cbs_ind].T;
		    }

		  /* Release pending requests (if any!) */
		  if(CBS_list.cbs[cbs_ind].n > 0)
		    { /* There is a pending request */
		      
		      cprintf("\n ftt_scheduler: pending requests");

		      /* Check if current budget enough to satisfy the request*/
		      if(CBS_list.cbs[cbs_ind].c < CBS_list.cbs[cbs_ind].dlc[0])
			{ /* Server budget not enough to serve the request */

			  cprintf("\n ftt_scheduler: budget not enough to serve - refill and update dedaln");
			  CBS_list.cbs[cbs_ind].d += CBS_list.cbs[cbs_ind].T; /* Update deadline */
			  CBS_list.cbs[cbs_ind].c += CBS_list.cbs[cbs_ind].Q; /* Refill server budget */
			}
		      
		      cprintf("\n ftt_scheduler: release CBS message ID=%x deadline=%d",\
			      CBS_list.cbs[cbs_ind].id,CBS_list.cbs[cbs_ind].d);

		      /* Release the request (put in EDF ready queue) */
		      new_rec.m_id       = CBS_list.cbs[cbs_ind].id;
		      new_rec.m_ssize    = (CBS_list.cbs[cbs_ind].dlc[0]*8*BIT_TIME_US)/SLOTV_US;
		      new_rec.m_deadline = CBS_list.cbs[cbs_ind].d;
		      EDF_RQ_insert(&EDF_rq, new_rec);
		      
		    }
		} /* If Served message found */
	    } /* For each CBS server */
	} /* For each served message */
      
      /* Clear the EC served messages queue */
      SCHED_CBS_nm=0; 
            
#endif /* CBS_SERVER */



#endif /* EDF_SCHED */



      /****************************************/
      /* Code common to EDF and RM schedulers */
      /****************************************/
      /* Check if any pending delete actions */
      /* Note: messages that have a request for removing and have already been activated */
      /*       (althought not scheduled) can be safely removed from the SRT.             */
      /*       This is signaled by the value 1 in the .sm_deleted field                  */
      if(del_pend == 1)
	{
	  for(mesg_ind=0 ; mesg_ind< SRDB_SRT_mesgtable.sm_nvars ; mesg_ind++) /* For each message in the SRT ...*/
	    {
	      /* If message can be safely removed ...  */
	      if(SRDB_SRT_mesgtable.sm_message[mesg_ind].sm_deleted == 2) 
		{
		  for(j=mesg_ind ; j < SRDB_SRT_mesgtable.sm_nvars-1 ; j++)
		    SRDB_SRT_mesgtable.sm_message[j] = SRDB_SRT_mesgtable.sm_message[j+1];
		  SRDB_SRT_mesgtable.sm_nvars--;
		} 
	    }

#ifdef ELASTIC_TM	  
	  /* If messages were removed, compress the resulting set */
	  ETM_compress(&SRDB_SRT_mesgtable,0);
#endif	     
	}



        /***************************************/
       /* Scheduling of asynchronous messages */
      /***************************************/
           	  	  
      /*******************************************************/
      /* The secheduling is performed in 3 steps:            */
      /*   1- First HARD messages are scheduled              */
      /*   2- Second SOFT messages are scheduled             */
      /*   3- Finally, UNCONSTRAINED messages are scheduled  */
      /*******************************************************/
	  
      /* First schedule HARD messages */ 
      for(mesg_ind = 0 ; mesg_ind < SRDB_ART_mesgtable.am_nvars; mesg_ind++) /* For each message in ART table ... */
	{
	  /* Scan ART table. Messages are ordered by decreasing priority, thus to build RM schedules */
	  /*   its enough to scan the table in ascending order */
	  
	  curmesg_a = &(SRDB_ART_mesgtable.am_message[mesg_ind]); /* Get copy of current message properties */
	  
	  if(curmesg_a->am_timeliness != TMLN_HARD)
	    continue; /* If not a HARD message, skip */
	  
	  if(curmesg_a->am_init <= 0 && ( sched_ec->am_load + curmesg_a->am_ssize*SLOTV_US ) < EC_MAX_us_AP )
	    { /* Message ready and fits within the Asynch window of the EC */
	      sched_ec->EC_AIDlist[sched_ec->am_nmesgs] = curmesg_a->am_id; /* Store ID */
	      sched_ec->am_nmesgs++; /* Update number of asynch. messages in EC */
	      sched_ec->am_load += (curmesg_a->am_ssize*SLOTV_US); /* Update the asynch load */
	      
	      /* Message scheduled. Update the next release instant */
	      curmesg_a->am_init += curmesg_a->am_period;
	    }
	}
      

      /* Next schedule SOFT messages */ 
      for(mesg_ind = 0 ; mesg_ind < SRDB_ART_mesgtable.am_nvars; mesg_ind++) /* For each message in ART table ... */
	{
	  /* Scan ART table. Messages are ordered by decreasing priority, thus to build RM schedules */
	  /*   its enough to scan the table in ascending order */
	      
	  curmesg_a = &(SRDB_ART_mesgtable.am_message[mesg_ind]); /* Get copy of current message properties */
	  
	  if(curmesg_a->am_timeliness != TMLN_SOFT)
	    continue; /* If not a SOFT message, skip */
	  
	  if(curmesg_a->am_init <= 0 && ( sched_ec->am_load + curmesg_a->am_ssize*SLOTV_US ) < EC_MAX_us_AP )
	    { /* Message ready and fits within the Asynch window of the EC */
	      sched_ec->EC_AIDlist[sched_ec->am_nmesgs] = curmesg_a->am_id; /* Store ID */
	      sched_ec->am_nmesgs++; /* Update number of asynch. messages in EC */
	      sched_ec->am_load += (curmesg_a->am_ssize*SLOTV_US); /* Update the asynch load */
	      
	      /* Message scheduled. Update the next release instant */
	      curmesg_a->am_init += curmesg_a->am_period;
	    }
	}

      
      /* Last schedule UNCONSTRAINED messages */ 
      for(mesg_ind = 0 ; mesg_ind < SRDB_ART_mesgtable.am_nvars; mesg_ind++) /* For each message in ART table ... */
	{
	  /* Scan ART table. Messages are ordered by decreasing priority, thus to build RM schedules */
	  /*   its enough to scan the table in ascending order */
	  
	  curmesg_a = &(SRDB_ART_mesgtable.am_message[mesg_ind]); /* Get copy of current message properties */
	  
	  if(curmesg_a->am_timeliness != TMLN_UNCONST)
	    continue; /* If not a UNCONSTRAINED message, skip */
	  
	  if(curmesg_a->am_init <= 0 && ( sched_ec->am_load + curmesg_a->am_ssize*SLOTV_US ) < EC_MAX_us_AP )
	    { /* Message ready and fits within the Asynch window of the EC */
	      sched_ec->EC_AIDlist[sched_ec->am_nmesgs] = curmesg_a->am_id; /* Store ID */
	      sched_ec->am_nmesgs++; /* Update number of asynch. messages in EC */
	      sched_ec->am_load += (curmesg_a->am_ssize*SLOTV_US); /* Update the asynch load */
	      
	      /* Message scheduled. Update the next release instant */
	      curmesg_a->am_init += curmesg_a->am_period;
	    }
	}


      /* Update the releases */
      for(mesg_ind = 0 ; mesg_ind < SRDB_ART_mesgtable.am_nvars; mesg_ind++) /* For each message in ART table ... */
	{
	  curmesg_a = &(SRDB_ART_mesgtable.am_message[mesg_ind]); /* Get copy of current message properties */
	  
	  //printf("\n Update de releases: mesg id=%d init=%d",curmesg->am_id,curmesg->am_init);
	  
	  /* Update the time until release */
	  curmesg_a->am_init--;
	  
	  /* Check for missed deadlines */
	  if(curmesg_a->am_init <= -(signed int)(curmesg_a->am_period))
	    {
	      switch(curmesg_a->am_timeliness)
		{
		case TMLN_HARD:
		  sched_ec->status=MISSED_DEADLINE;
		  sched_ec->err_code=SCH_ERRCODE_AM_HARD_MISS_DEADLINE;
		  cprintf("\n ! Scheduler fatal error: Missed deadline in Asynchronous Hard message (ID=%x)",curmesg_a->am_id);
		  break;
		  
		case TMLN_SOFT:
		  sched_ec->err_code=SCH_ERRCODE_AM_SOFT_MISS_DEADLINE;
		  break;
		  
		case TMLN_UNCONST:
		  sched_ec->err_code=SCH_ERRCODE_AM_UNCON_MISS_DEADLINE;
		  break;
		}
	    }
	  
	} /* For each message (update releases) */
      

      /* If plan sucessfully built, set plan status flag */
      if(sched_ec->status == BUSY_SCHED)
	sched_ec->status = UPDATED;

      /* Exit mutex region - scheduler is finished, so other process can access the SRT */
      mutex_unlock(&srt_access_mutex);

      /* Wait for next invocation */
      task_endcycle();
    }
}



signed char SRDB_printsched(EcType *ec)
/***********************************************************************/
/*!
\brief  This function prints the EC-schedule contents

 This function prints the EC-schedule contents 

\param ec: pointer to EC schedule
     
\returns error code\n
					0 - OK\n
					1 - Schedule empty!

***********************************************************************/
{
  int i; /* General counter */
    
  cprintf("\n EC:  #mesgs(sync= %3d async=%3d) ; load(us)(sync=%5lu async=%5lu) ",ec->sm_nmesgs,ec->am_nmesgs, \
	  ec->sm_load, ec->am_load);
  
  cprintf("\n    Sync.  mesgs:");
  for(i=0;i < ec->sm_nmesgs;i++)
    {
      cprintf(" [%x/%d]",ec->EC_SIDlist[i],ec->EC_SSizelist[i]);
    }

  cprintf("\n    Async. mesgs:");
  for(i=0;i < ec->am_nmesgs;i++)
    {
      cprintf(" [%x]",ec->EC_AIDlist[i]);
    }
  
  cprintf("\n");
  
  return 0;
}


TASK ftt_dispatcher(void *arg)
/****************************************************************************/
/*!
\brief This task is invoqued periodically (at the beginning of each EC)

	This task is invoqued periodically (at the beginning of each EC).\n

	- scans current EC (sched_ec),\n
	- builds respective master message and,\n
	- send it to the bus.

*****************************************************************************/
{
  /****************/
  /* Declarations */
  /****************/

  unsigned char i;

  ETH_ADDR broadcast; /* Ethernet addresses */
  void *TxBuff;       /* Pointer to a Eth Transmission buffer */
  FTT_TM_PKT *tm_pkt; /* Pointer to TM contents */
  WORD *dbgw_tm_pkt;  /* Pointer to TM contents (for debug) */

  int pack_index;     /* Auxiliary Counter used to store data in TM data field */
  int tm_datalength;    /* Data lenght of the TM frame */


#ifdef ETM_MEASURE
  TIME ec_inst; /* Instant of tx of the TM (in usec) */ 
#endif

  /********/
  /* Code */
  /********/

  //kern_printf("DISP_INIT");
  eth_str2Addr("FF:FF:FF:FF:FF:FF",&broadcast);


  if( (TxBuff = netbuff_get(&Pool_TxBuff,NON_BLOCK)) == NULL)
    {
      perror("[ftt_dispatcher] Error! netbuff_get() returned null ");
      sys_end();
    }
  
  while(1)
    {

      //kern_printf("DISP_INST");
      
      /********************/
      /* Check  EC status */
      /********************/
      switch( disp_ec->status )
	{
	case OUT_OF_DATE:        /* Starting up :  do nothing */
	  break;  

	case MISSED_DEADLINE:    /* Missed deadline detected by scheduler, terminate */
	  perror("[ftt_dispatcher] Fatal error: missed deadline in synchronous message");
	  sys_end();
	  
	case BUSY_SCHED:         /* Scheduler did not build plan in time, terminate */
	case DISPATCHED:         /* Scheduler has not been started                  */
	  perror("[ftt_dispatcher] Fatal Error : Scheduler has not finished in time");
	  sys_end();

	case UPDATED:            /* OK. Follow down to normal plan dispatching */
	  break;
	  
	default:
	  perror("[ftt_dispatcher] Fatal Error : Invalid EC Status (Strange error!)");
	  sys_end();
	  

	} /*Switch */


      /**********************************/
      /* Build trigger message (TM)     */
      /**********************************/
      eth_setHeader(TxBuff,broadcast,ETH_FTT_TYPE);

#ifdef ETM_MEASURE
      ec_inst=sys_gettime(NULL);
#endif

      tm_pkt= (FTT_TM_PKT *)eth_getFDB(TxBuff);
      
      /* DEBUG */
      dbgw_tm_pkt = (WORD *)tm_pkt;

      /* Build TM Header contents : TM_ID + TM_Flags + nsmesgs + namesgs */
      tm_pkt->TM_Header.TM_id    = htons((WORD)MST_MESG_ID);
      tm_pkt->TM_Header.TM_flags = htons((WORD)(TMSEQN_MASK & tm_seqn));
      tm_pkt->TM_Header.TM_nsm   = htons((WORD)disp_ec->sm_nmesgs);
      tm_pkt->TM_Header.TM_nam   = htons((WORD)disp_ec->am_nmesgs);
      //cprintf("\n TM: nsmesgs=%d namesgs=%d",disp_ec->sm_nmesgs,disp_ec->am_nmesgs);
      

      /***********************/
      /* Build TM Data field */
      /***********************/
      /* Synchronous message data */
      pack_index=0;
      
      for( i=0; i < disp_ec->sm_nmesgs; i++ )
	{
	  tm_pkt->TM_Data[pack_index++] = (BYTE)(disp_ec->EC_SIDlist[i] >> 8);
	  tm_pkt->TM_Data[pack_index++] = (BYTE)(disp_ec->EC_SIDlist[i]);
	  tm_pkt->TM_Data[pack_index++] = (BYTE)(disp_ec->EC_SSizelist[i]);
        }
      
      /* Asynchronous message data */
      for( i=0; i < disp_ec->am_nmesgs; i++ )
	{
	  tm_pkt->TM_Data[pack_index++] = (BYTE)(disp_ec->EC_AIDlist[i] >> 8);
	  tm_pkt->TM_Data[pack_index++] = (BYTE)(disp_ec->EC_AIDlist[i]);
        }
 
      /*******************************************************************************/
      /* Send packet. Note that only the data field len of the Eth frame is supplied */
      /*  eth_sendPkt() function adds the len of addresses and frame type fields     */
      /*******************************************************************************/
      tm_datalength=sizeof(FTT_TM_HEADER) + disp_ec->sm_nmesgs*3*sizeof(BYTE)+ disp_ec->am_nmesgs*2*sizeof(BYTE);
      
      /* DEBUG - TM contents */
      //cprintf("... Len:%d\n  ",tm_datalength);
      //for(i=0;i<(tm_datalength+12);i++)
      //cprintf("%x:",*((BYTE *)(TxBuff+i)));
      //cprintf("\n ID=%x Flags=%x nsm=%x nam=%x",ntohs((WORD)(*(dbgw_tm_pkt))),ntohs((WORD)(*(dbgw_tm_pkt+1))),ntohs((WORD)(*(dbgw_tm_pkt+2))),ntohs((WORD)(*(dbgw_tm_pkt+3))));
      // DEBUG - End */

      eth_sendPkt(TxBuff, sizeof(FTT_TM_HEADER) + disp_ec->sm_nmesgs*3*sizeof(BYTE)+ \
		  disp_ec->am_nmesgs*2*sizeof(BYTE));
      

      tm_seqn++; /* Update TM sequence number */

      /***************/
      /* Swap ECs    */
      /***************/
      disp_ec->status = DISPATCHED;
      temp_ec = sched_ec;
      sched_ec = disp_ec;
      disp_ec = temp_ec;  
      
      
      task_activate(pid_ftt_scheduler); /* Activate ftt_pscheduler() */
        
      task_endcycle(); /* Wait for next EC */
    }

}


void ftt_ethproc(void *eth_pkt)
/****************************************************************************/
/*!
\brief This function is triggered by the reception of FTT Ethernet Frame.

  This function is triggered by the reception of FTT Ethernet Frame.\n   
  Depending on the frame type several actions are performed             
                                                                         
****************************************************************************/
{

  /****************/
  /* Declarations */
  /****************/
  FTT_TM_PKT *ftt_tm_pkt;
  WORD frame_type,frame_id,nsm,nam;
  
  //BYTE *ftt_pkt_byte; /* Pointers to Ethernet frame received */ 
  WORD *ftt_pkt_word;

  short int frame_t2deadline; /* Time to deadline carried in the message received (for data messages) */
  WORD frame_flags;      /* Message flags data field */

  WORD sig_type;  /* Signal type*/

#ifdef CBS_SERVER
  /* SIGNAL+CBS related vars */
  WORD cbs_req_id, cbs_req_len; /* id and dlc of the request to the CBS server */
#endif

  /********/
  /* Code */
  /********/

  /* Get pointer to frame just received */
  ftt_pkt_word=(WORD *)eth_getFDB(eth_pkt);
  //ftt_pkt_byte=(BYTE *)eth_getFDB(eth_pkt);

  /* Get ftt frame type */
  frame_type=(ntohs(*ftt_pkt_word) & MESGTYPE_MASK);
  
  switch(frame_type)
    {
      
    case MST_MESG_ID:  
      /*********************************/
      /***** Trigger Message received? */
      /*********************************/      
     
      /* Get pointer to TM data contents */
      ftt_tm_pkt=(FTT_TM_PKT *)eth_getFDB(eth_pkt);
      
      /* Get number of sync. and asynch. messages in the EC */
      nsm=ntohs(ftt_tm_pkt->TM_Header.TM_nsm);
      nam=ntohs(ftt_tm_pkt->TM_Header.TM_nam);

      /* Message processing */
      cprintf("\n Trigger Message received");
  
      /* End of processing */
      break;
    
      
    case SDATA_MESG_ID: 
      /******************************************/
      /***** Synchronous Data Message received  */
      /******************************************/

      /* Get ftt frame header info */
      /* Synch Data Header ID(WORD) + Flags(WORD) + t2deadline(WORD) */ 
      frame_id         = ntohs(*ftt_pkt_word); 
      frame_flags      = ntohs(*(ftt_pkt_word+1));
      frame_t2deadline = ntohs(*(ftt_pkt_word+2));
            
      /* Message processing */
      //cprintf("\n Synchronous Data message received");
      
      
      /* End of processing */
      break;


    
        /******************************************/
       /**** Asynchronous Data Message received  */
      /******************************************/
    case ADATA_MESG_ID_DATA_SHORT: 
      break;

    case ADATA_MESG_ID_ALARM:
      break;

    case ADATA_MESG_ID_SIGNAL:
      /* Get ftt frame header info */
      /* Asynch Data Header ID(WORD) + Flags(WORD) + t2deadline(WORD) */ 
      frame_id         = ntohs(*ftt_pkt_word); 
      frame_flags      = ntohs(*(ftt_pkt_word+1));
      frame_t2deadline = ntohs(*(ftt_pkt_word+2));

      /* Get signal type (1st byte after the header) */
      sig_type = ntohs(*(ftt_pkt_word+3));
          
      cprintf("\n SIGNAL received (%x)",sig_type);
      
      /* Select the indicated signal */
      switch(sig_type)
	{
	case SIGNAL_CBS_REQ_TYPE:

#ifdef CBS_SERVER
	  cbs_req_id  = ntohs(*(ftt_pkt_word+4));
	  cbs_req_len = ntohs(*(ftt_pkt_word+5));

	  cprintf(" [SIGNAL_CBS_REQ ID=%x dlc=%u]",cbs_req_id,cbs_req_len);

	  CBS_request(&CBS_list, cbs_req_id, cbs_req_len);
#endif	  	  
	  break;
	  
	}
                  
      /* End of processing */
      break;

    case ADATA_MESG_ID_DATA_LONG:
      break;

    } /* End Switch(frametype) */
  
}



/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/



    /*****************************************************/
   /*****************************************************/
  /*    General functions                              */
 /*****************************************************/
/*****************************************************/

#ifdef ETM_MEASURE

void save_etmtstres(void *arg)
/********************************************************************/
/* This function is called when the system returns to real mode     */
/* (via sys_atrunlevel(). It saves the contents of a data structure */
/* on a file.                                                       */
/********************************************************************/
{

    DOS_FILE *f;
    int err;
    unsigned char linfo[80];
    WORD eid,eev;
   
    cprintf("\n\n Saving results file ....\n");

    f=DOS_fopen("etmrese.txt","w");
    if(!f)
      {
	err=DOS_error();
	cprintf("\n Error opening data file(1)! ");
	cprintf("\n Error code=%d",err);
	sys_end();
      }
    
    for(eid=0;eid<MAX_VAR;eid++)
      {
	sprintf(linfo,"\nID %x",eid);
	DOS_fwrite(linfo,sizeof(char),strlen(linfo),f);
	for(eev=0;eev<EVT_TAB_maxevt;eev++)
	  {
	    sprintf(linfo," %lu",evt_tab[eid][eev]);
	    DOS_fwrite(linfo,sizeof(char),strlen(linfo),f);
	    if(evt_tab[eid][eev]==0) break;
	  } 
      }
    
    DOS_fclose(f);

    f=DOS_fopen("etmresa.txt","w");
    if(!f)
      {
	err=DOS_error();
	cprintf("\n Error opening data file(2)! ");
	cprintf("\n Error code=%d",err);
	sys_end();
      }
    
    for(eev=0;eev<etoact;eev++)
      {
	sprintf(linfo,"\nID %x OP %u Ti %lu",evt_act[eev].id,evt_act[eev].action,evt_act[eev].inst);
	DOS_fwrite(linfo,sizeof(char),strlen(linfo),f);
      }
    
    DOS_fclose(f);
    

    cprintf("\n  Results saved. Bye!");
}
#endif    




/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/



    /*****************************************************/
   /*****************************************************/
  /*    Boot functions                                 */
 /*****************************************************/
/*****************************************************/

int ftt_minit()
/*************************************************************************/
/*! 
\brief This function is responsible for initializing the ftt system
                                                                     
    This function is responsible for initializing the ftt system:\n      
            - Network set-up\n                                           
            - System tasks creation (scheduler, dispatcher, ...)\n       
            - Data structures initialization\n                           
                                                                       
   and MUST BE CALLED BEFORE ANYTHING ELSE                             
*************************************************************************/
{

  /*************************/
  /* Variable declarations */
  /*************************/
  HARD_TASK_MODEL ftt_dispatcher_tm,ftt_scheduler_tm;

  NOP_mutexattr_t srt_access_mutex_m; /* Mutex to serialize access to SRT between */
                                      /* all the tasks that use it (NO Protocol)  */ 


    /*******************/
   /* Code            */
  /******************/

  //set_exchandler_grx(); /* After execption prints PID ... */

    /************************/
   /* Initialize NOP mutex */
  /************************/
  NOP_mutexattr_default(srt_access_mutex_m);
  mutex_init(&srt_access_mutex,&srt_access_mutex_m);

    /**********************/
   /* Net initialization */
  /**********************/
  netbuff_init(&Pool_TxBuff,2,ETH_MAXLEN_BYTES); /* Allocate Tx buffers */
  eth_setProtocol(ETH_FTT_TYPE,ftt_ethproc);	/* Bind frame type Rxption with handler */

  /*********************************************/
  /* <<<<< Module organization >>>>>           */
  /* Note : there are the following modules:   */
  /*         Level 0 : EDF                     */
  /*         Level 1 : CBS (Guest on level 0)  */
  /*         Level 2 : EDF                     */
  /*         Level 3 : CBS (Guest on level 1)  */
  /*         Level 4 : RR                      */
  /*   -The dispatcher is registered at level 0 (Highest priority level) */
  /*   -The eth_init tasks at level 1 (2nd highest priority )            */
  /*   -ALL OTHER HRT TASKS at level 2                                   */
  /*   -ALL OTHER SRT TASKS at level 3                                   */
  /*   This way the dispatcher will allways preempt any other tasks when it becomes ready */
  /*      providing a more regular Trig message. The taks related with ethernet get the   */
  /*      following level of priority.         */
  /*********************************************/

    /******************************/
   /* Create ftt_dispatcher task */
  /******************************/
  disp_mit=(TIME)EC_LEN;
  disp_wcet=(TIME)disp_mit-1000;
  hard_task_default_model(ftt_dispatcher_tm);
  hard_task_def_mit(ftt_dispatcher_tm,disp_mit);
  hard_task_def_wcet(ftt_dispatcher_tm,disp_wcet);
  hard_task_def_periodic(ftt_dispatcher_tm);
  hard_task_def_level(ftt_dispatcher_tm,0); /* Register task at level 0 */
  
  if ((pid_ftt_dispatcher=task_create("ftt_dispatcher", ftt_dispatcher, &ftt_dispatcher_tm, NULL)) == -1)
    {
      perror("[FTTETM1.c(main)] Could not create <ftt_dispatcher>:");
      sys_end();
    }
  perror("\n[FTTETM1.c(main)] <ftt_dispatcher> task created:");

    /******************************/
   /* Create ftt_scheduler  task */
  /******************************/
  /* Important note: since the dispatcher can have some jitter when activating the schduler */
  /*  due to activation jitter and itself execution jitter, the mit should be reduced for */
  /*  an amount greatest than the max jitter, to avoid activation before mit */  
  sched_mit=(TIME)EC_LEN-2000;
  sched_wcet=sched_mit-1000; 
  hard_task_default_model(ftt_scheduler_tm);
  hard_task_def_mit(ftt_scheduler_tm,sched_mit);
  hard_task_def_wcet(ftt_scheduler_tm,sched_wcet);
  hard_task_def_aperiodic(ftt_scheduler_tm);
  hard_task_def_level(ftt_scheduler_tm,2); /* Register task at level 2 */
  
  if ((pid_ftt_scheduler=task_create("ftt_scheduler", ftt_scheduler, &ftt_scheduler_tm, NULL)) == -1)
    {
      perror("[FTTETM1.c(main)] Could not create <ftt_scheduler>:");
      sys_end();
    }
  perror("\n[FTTETM1.c(main)] <ftt_scheduler> task created:");
   

    /********************/
   /* Initializations  */
  /********************/

  tm_seqn=0; /* TM sequence number */
	
  /* Init mesg table */
  SRDB_SRT_mesgtable.sm_nvars=0; /* SRT empty */
  SRDB_ART_mesgtable.am_nvars=0; /* ART empty */

  /* Init plans and pointers to plans */
  EC_1.status = OUT_OF_DATE;
  EC_2.status = OUT_OF_DATE;
  sched_ec=&EC_1;
  disp_ec=&EC_2;

#ifdef CBS_SERVER
  CBS_init(&CBS_list);
#endif
	
  cprintf("\n Starting FTT Ethernet Master 1 ...\n");

#ifdef ETM_MEASURE
  for(eid=0;eid<MAX_VAR;eid++)
    {
      evt_vec[eid]=0; /* Number of events */
      for(eev=0;eev<EVT_TAB_maxevt;eev++) /* Events table */
	evt_tab[eid][eev]=0;
    }
  etoact=0; /* Total of actions */
#endif	


  cprintf("\n\n System parameters:");
#ifdef RM_SCHED
  cprintf("\n    >Rate Monotonic (RM) scheduler");
#endif
#ifdef EDF_SCHED
  cprintf("\n    >Earliest Deadline First (EDF) scheduler");
#endif
#ifdef ELASTIC_TM
  cprintf("\n    >Using the Elastic Task Model!!");
#endif
#ifdef CBS_SERVER
  cprintf("\n    >CBS servers available !!");
#endif
  cprintf("\n    ----------------------------------------------------");
  cprintf("\n    > EC Length:%ld(ms)",EC_LEN/1000);
  cprintf("\n    > Bits in each EC:%ld",EC_LEN_bits);
  cprintf("\n    > TM length (worst case, bits)=%ld",TM_LEN_bits);
  cprintf("\n    > Max bits of synch.  window:%ld",EC_MAX_bits_SP);
  cprintf("\n    > Max bits of asynch. window:%ld",EC_MAX_bits_AP);
  cprintf("\n    ----------------------------------------------------");
  //cprintf("Any key to continue");
  //keyb_getchar();

  return 0;
}



int ftt_mstart()
/*************************************************************************/
/*! 
\brief This function is responsible for starting the system 

    This function is responsible for starting the system.\n              
    It can be called before or after the message definition.\n            
    If called after the task definition the system starts with the same time reference
	for all the messages, wich allows to use the init field to set relative phasing.\n                   
   Note: ftt_init must be called before this function                  
                                                                       
*************************************************************************/
{

  struct timespec twait;

    /*************************/
   /* Build first two plans */
  /*************************/

  cprintf("\n Creating first two plans...");
  task_activate(pid_ftt_scheduler); /* ftt_scheduler fills sched_plan */

  // task_delay(sched_mit);  

  twait.tv_sec=0;
  twait.tv_nsec=sched_mit*1000;
 
  nanosleep(&twait,NULL);


#if 0
  /* Print the built schedule */
  SRDB_printsched(sched_ec);
  cprintf("Any key to continue");
  keyb_getchar();
#endif

  /* Swap plans */
  temp_ec = sched_ec;
  sched_ec = disp_ec;
  disp_ec = temp_ec;
  
  task_activate(pid_ftt_scheduler);
  //task_delay(sched_mit);
 
  nanosleep(&twait,NULL);

#if 0
  /* Print the built schedule */
  SRDB_printsched(sched_ec);
  cprintf("Any key to continue");
  keyb_getchar();
#endif
 
  /* Now both disp & sched ECs are built */
  /* From now on, when the dispatcher finishes disp_ec, it switches */
  /* disp_ec and sched_ec and actiates the scheduler */
  
  /*******************/
  /* Activate system */
  /*******************/
  cprintf("\n Boot OK! Starting system...\n");
  task_activate(pid_ftt_dispatcher);
  cprintf("\n Dispatcher is now working...");

  return 0;
}

/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< END >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/

