/*********************************************************************************************
 ********************************************************************************************/
/*! \file ftts1_0.h
    \brief Macros, global vars and function prototypes related with the FTT-Ethernet\n
              Slave implementation 

 Macros, global vars and function prototypes related with the FTT-Ethernet\n
              Slave implementation

 Class: FTT-Ethernet\n                                  
 Type:  FTT Ethernet Slave code\n
 Platform: Shark\n
 Name: ftts1_0.h\n
 
 By: Paulo Pedreiras\n
 Date: 2003-Aug-4
 
 ***********************************************************************************************
 ***********************************************************************************************/

/********************************/
/* Includes 			*/
/********************************/

#include <stdlib.h>
#include <kernel/kern.h>
#include <modules/sem.h>
#include <kernel/int_sem.h>
#include <modules/hartport.h>
#include <modules/cabs.h>
#include <drivers/keyb.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "kernel/mem.h"
#include "ll/i386/cons.h"

#include "drivers/udpip.h"
#include "drivers/eth.h"
#include "netbuff.h"
#include "fttps1.h"


#ifndef __FTT_SL_HEADER__
#define __FTT_SL_HEADER__

/**********************************/
/* Macros                         */
/**********************************/

/*!
\brief Message properties
*/
#define SET_MESG_PROP(mp,mid,msize,mdeadline,mtype,mstatus,mpidcallback)\
       (mp)->m_id           = (mid),       \
       (mp)->m_size         = (msize),     \
       (mp)->m_deadline     = (mdeadline), \
       (mp)->m_type         = (mtype),     \
       (mp)->m_status       = (mstatus),   \
       (mp)->m_cbs_smid     = 0,           \
       (mp)->m_pid_callback = (mpidcallback)

#ifdef CBS_SERVER
#define SET_CBS_SIG_ID(mp,ms_id)\
       (mp)->m_cbs_smid     = (ms_id)

#endif

/********************************/
/* Global declarations          */
/********************************/

/*!
\brief EC length (us)
*/
unsigned long EC_LEN;  

/*!
\brief ProcID of the task responsible by producing the sync. mesgs
*/
PID pid_ftt_sdispatcher;        

/*!
\brief Table with node's message properties
*/
NRDB_tabtype NRDB_mesgtable;	

/*!
\brief CABS used for holding data, one for each synchronous message handled by the station
*/
CAB cabset[STA_MAX_VAR];

/*!
\brief Com ports used for holding asynchronous messages data
*/
PORT portset[STA_MAX_VAR];      

/*!
\brief Pool of Eth TX buffers
*/
struct netbuff Pool_TxBuff;     

/*!
\brief Instant of reception of last trigger message
*/
struct timespec tm_timestamp;  

/*!
\brief Table with the messages scheduled to be produced within the EC
*/
STA_M2P_tabtype mesg2p_table;  


/*!
\brief Buffer used by send()
*/
BYTE SEND_mesgbuf[1600];

/*!
\brief Buffer used by receive()
*/
BYTE RECEIVE_mesgbuf[1600];

/*!
\brief Buffer used by send()
*/
BYTE FTT_SEND_mesgbuf[1600];

/*!
\brief Buffer used by receive()
*/
BYTE FTT_RECEIVE_mesgbuf[1600];


/* Debug variables */

/*!
\brief Incremented by each message transmission (Debug variable)
*/
int dbg_txcnt; 

/*!
\brief Incremented by each TM received (Debug variable)
*/
int dbg_eccnt; 


/* Used for debuging actiavtion error */

/*!
\brief Used for debuging actiavtion error
*/
struct timespec smp_activation;

/*!
\brief Used for debuging actiavtion error
*/
long maxdif;

/*!
\brief Used for debuging actiavtion error
*/
long mindif;

/*!
\brief Used for debuging actiavtion error
*/
int prim_sample;

/*!
\brief Used for debuging actiavtion error
*/
int txact_late;

/*!
\brief Used for debuging actiavtion error
*/
int txact_jit;

/*!
\brief Used for debuging actiavtion error
*/
int txact_early;
/*******/

/*!
\brief Sem for controling ftt_sdispatcher execution
*/
internal_sem_t sem; 



    /*****************************************************/
   /*****************************************************/
  /*    NRDB handling functions                        */
 /*****************************************************/
/*****************************************************/


signed char NRDB_addmesg(NRDB_mesgtype *aux_var);/* Inserts a variable in the station variables' table */
signed char NRDB_delmesg(WORD MesgID, WORD mesg_type);/*  Removes a message from the station variables' table */
signed char NRDB_changemesg(WORD MesgID, WORD mesg_type, WORD size, WORD deadline); /*  Removes a message from the station variables' table */
int NRDB_printmesg(void); /* Prints info about the messages locally configured (both consumed and received) */
signed char NRDB_getstatus(unsigned int MesgID, WORD mesg_type,NRDB_mesgtype *mesg_prop); /* Returns variable's status byte of message m_id */


    /*****************************************************/
   /*****************************************************/
  /*    Communication primitives                       */
 /*****************************************************/
/*****************************************************/

signed char produce (unsigned int MesgID, unsigned char *mesg_value);
        /* Copies a message's data into local buffer */
signed char consume (unsigned int MesgID, unsigned char *mesg_value, long *deadline); 
        /* Copies a message's value from local buffer the value buffer in the Sta. Message Table */
int ftt_send(unsigned int MesgID, WORD mesg_type, unsigned char *mesg_value, WORD mesg_dlc); 
        /* Checks if station is producer of this var. Places the data in a local fifo */
int send(unsigned int MesgID, WORD mesg_type, unsigned char *mesg_value); 
        /* Interface for ftt_send(). Used to send regular asynch. messages */
int send_l (unsigned int MesgID, WORD mesg_type, unsigned char *mesg_value, WORD mesg_dlc);
        /*  Sends long asynchronous messages using CBS servers */
int ftt_receive (unsigned int MesgID, WORD mesg_type,unsigned char *mesg_value,WORD *mesg_dlc, long *deadline);
        /* Checks if station is receiver of this var.  Gets data from a receiving fifo deadline info */
int receive (unsigned int MesgID, WORD mesg_type,unsigned char *mesg_value,long *deadline);
        /* Consumes regular asynch. messages */
int receive_l (unsigned int MesgID, WORD mesg_type, unsigned char *mesg_value, WORD *mesg_dlc, long *deadline);
        /* Consumes long asynch. messages */


    /*****************************************************/
   /*****************************************************/
  /*    FTT-protocol  functions                        */
 /*****************************************************/
/*****************************************************/


signed char ftt_get2tx_sdata(unsigned int MesgID, unsigned char *mesg_value, WORD *sdm_len, \
			     BYTE *mesg_seqn, short int *mesg_deadline, PID *pid_callback);
        /*  Gets the most recent variable's value from the respective CAB */
signed char ftt_rx2update_sdata(unsigned int MesgID, unsigned char *mesg_value, \
				WORD mesg_flags, short int mesg_t2deadline);
        /* Updates the CAB contents relative to a specific message with data passed as argument */
signed char ftt_get2tx_adata(unsigned int MesgID, unsigned char *mesg_value, WORD *adm_len, \
			     BYTE *mesg_seqn, short int *mesg_deadline, PID *pid_callback);
        /* Gets the most recent variable's value from the respective FIFO */
signed char ftt_rx2update_adata(unsigned int MesgID, unsigned char *mesg_value, \
				WORD mesg_flags, short int mesg_t2deadline);
        /* Updates the FIFO contents relative to a specific asynchronous message */
signed char ftt_test_producer(unsigned int MesgID);
        /* Tests if a station is a producer of the message with the ID passed in the arguments */
signed char ftt_test_consumer(unsigned int MesgID);
        /* Tests if a station is a consumer of the message with the ID passed in the arguments*/


  /************************************************************************/
 /*                     FTT-Ethernet Tasks                               */
/************************************************************************/

void ftt_ethproc(void *eth_pkt);
        /* Handles incoming Ethernet Frames.*/

TASK ftt_sdispatcher(void *arg);
        /* This task is responsible by timely produce the station messages.*/


    /*****************************************************/
   /*****************************************************/
  /*           ftt boot functions                      */
 /*****************************************************/
/*****************************************************/

int ftt_ninit();
        /* Initializes FTT system: System tasks creation and Data structures initialization */
int ftt_nstart();
/*  System start (nodes). It can be called before or after the message definition */

#endif
