/*********************************************************************************************
 ********************************************************************************************/

/*! \file fttsnif.c
\brief Sniffer application

 Sniffer application\n
 Class: FTT-Ethernet\n                                  
 Type:  FTT Ethernet Slave code\n
 Platform: Shark\n
 Name: fttsnif.c\n
 
 By: Paulo Pedreiras\n
 Date: 2001-Oct\n
 
 ***********************************************************************************************
 ***********************************************************************************************/

/********************************/
/* Includes 			*/
/********************************/

#include <stdlib.h>
#include <stdio.h>
#include <kernel/kern.h>
#include <modules/sem.h>
#include <modules/hartport.h>
#include <modules/cabs.h>
#include <drivers/keyb.h>
#include <string.h>
#include <time.h>

#include "kernel/mem.h"
#include "ll/i386/cons.h"

#include "drivers/udpip.h"
#include "drivers/eth.h"
#include "netbuff.h"

#include "fttps1.h"

/********************************/
/* Global declarations */
/********************************/

/*! \def SNIFFVER
\brief Sniffer version
*/
#define SNIFFVER 1.0

unsigned char mostra; 


void ftt_ethproc(void *eth_pkt)
/****************************************************************************/
/*! 
\brief This function waits for the reception of an Ethernet Frame, and prints its contents

This function waits for the reception of an Ethernet Frame, and prints its contents.

*****************************************************************************/

{

  /****************/
  /* Declarations */
  /****************/
  WORD *ftt_pkt_word;
  BYTE *ftt_pkt_byte;
  WORD frameid,frametype,tm_flags,tm_nsm,tm_nam;
  WORD sdataid,adataid;
  BYTE sdatalen;
  
  WORD dmesgh_id,dmesgh_flags;
  short int dmesgh_deadline;
  BYTE dmesg_data[3];
  int i;

  TIME tevent;
  
  /********/
  /* Code */
  /********/

  //cprintf(".");


  if(!mostra) return;

  /* Get time of the event */
  tevent = sys_gettime(NULL);
  
  /* Get pointers to frame just received */
  ftt_pkt_word=(WORD *)eth_getFDB(eth_pkt);
  ftt_pkt_byte=(BYTE *)eth_getFDB(eth_pkt);

  /* Get ftt frame ID */
  frameid=ntohs(*ftt_pkt_word);

  frametype=frameid & MESGTYPE_MASK;
  
  switch(frametype)
    {
    case MST_MESG_ID:
      
      /* DEBUG- printf received raw data (note: words in network format */
      //for(i=0;i<14;i++)
      //cprintf("%x:",(BYTE)( *(ftt_pkt_byte + i) ) );
      
      
      /* Get number of sync. messages in the EC */
      tm_flags=ntohs(*(ftt_pkt_word+1));
      tm_nsm=ntohs(*(ftt_pkt_word+2));
      tm_nam=ntohs(*(ftt_pkt_word+3));


      cprintf("\n>>>[TM] Rx_Time=%lu ID=%x Seqn=%u #sm=%u #am=%u",tevent,frameid,(TMSEQN_MASK & tm_flags),tm_nsm,tm_nam);
      
      cprintf("\n  SM [ID Len]:");
      ftt_pkt_byte += 4*sizeof(WORD);
      for(i=0;i<tm_nsm;i++)
	{
	  sdataid=(*ftt_pkt_byte)<<8|(*(ftt_pkt_byte+1));
	  sdatalen=*(ftt_pkt_byte+2);
	  cprintf(" [%x %u]",sdataid,sdatalen);
	  ftt_pkt_byte+=3*sizeof(BYTE);
	 
	}

      cprintf(" AM [ID]:");
      for(i=0;i<tm_nam;i++)
	{
	  adataid=(*ftt_pkt_byte)<<8|(*(ftt_pkt_byte+1));
	  cprintf(" [%x]",adataid);
	  ftt_pkt_byte+=2*sizeof(BYTE);
	 
	}

      
      break;

    case SDATA_MESG_ID:
      /* Get data message header */
      dmesgh_id=ntohs(*(ftt_pkt_word));
      dmesgh_flags=ntohs(*(ftt_pkt_word+1));
      dmesgh_deadline=(signed short int)ntohs(*(ftt_pkt_word+2));
      
      /* Get first data bytes */
      ftt_pkt_byte+=sizeof(FTT_SDM_HEADER);
      dmesg_data[0]=*ftt_pkt_byte;
      dmesg_data[1]=*(ftt_pkt_byte+1);
      dmesg_data[2]=*(ftt_pkt_byte+2);

      /* Print info on screen*/
      cprintf("\n[SDM] Rx_Time=%lu ID=%x Flags=%u Deadline(DDL_RES_US)=%d Data %2x %2x %2x",\
	      tevent,dmesgh_id,dmesgh_flags,dmesgh_deadline,dmesg_data[0],dmesg_data[1],dmesg_data[2]);
      
      break;

    case ADATA_MESG_ID_DATA_SHORT:
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:
    case ADATA_MESG_ID_DATA_LONG:
      /* Get data message header */
      dmesgh_id=ntohs(*(ftt_pkt_word));
      dmesgh_flags=ntohs(*(ftt_pkt_word+1));
      dmesgh_deadline=(signed short int)ntohs(*(ftt_pkt_word+2));
      
      /* Get first data bytes */
      ftt_pkt_byte+=sizeof(FTT_SDM_HEADER);
      dmesg_data[0]=*ftt_pkt_byte;
      dmesg_data[1]=*(ftt_pkt_byte+1);
      dmesg_data[2]=*(ftt_pkt_byte+2);

      /* Print info on screen*/
      cprintf("\n[ADM] Rx_Time=%lu ID=%x Flags=%u Deadline(DDL_RES_US)=%d Data %2x %2x %2x",\
	      tevent,dmesgh_id,dmesgh_flags,dmesgh_deadline,dmesg_data[0],dmesg_data[1],dmesg_data[2]);
      
      break;

      

    }  
  /* End of processing */
  return;
    

}
 

/********************************/
/*! 
\brief Main function

Main function.
********************************/

int main(int argc, char **argv)
{
  
  int i;

  mostra = 1;

  cprintf("\n **** FTT Sniffer V%2.2f ****\n",SNIFFVER);
  cprintf("\n SPACE to pause/resume");
  cprintf("\n CTRL+C to quit\n");
  /*************/
  /* Setup net */
  /*************/
  eth_setProtocol(ETH_FTT_TYPE,ftt_ethproc);	/* Bind frame type Rxption with handler */
 
  /*********************/
  /* Keep main running */
  /*********************/
  while(1)
    {
      if((i=keyb_getch(NON_BLOCK)))
	if(i==' ')
	  {
	  if(mostra == 0)
	    mostra = 1;
	  else
	    mostra = 0;
	  }
    }
  
  return 0;

}

